# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=no-else-return

"""Module for working with Results."""

import copy
import numpy
from ._qiskiterror import QISKitError
from ._quantumcircuit import QuantumCircuit


class Result(object):
    """ Result Class.

    Class internal properties.

    Methods to process the quantum program after it has been run

    Internal::

        result = {
            "job_id": --job-id (string),
                      #This string links the result with the job that computes it,
                      #it should be issued by the backend it is run on.
            "status": --status (string),
            "result":
                [
                    {
                    "data":
                        {  #### DATA CAN BE A DIFFERENT DICTIONARY FOR EACH BACKEND ####
                        "counts": {'00000': XXXX, '00001': XXXXX},
                        "time"  : xx.xxxxxxxx
                        },
                    "status": --status (string)--
                    },
                    ...
                ]
            }
    """

    def __init__(self, qobj_result):
        self._result = qobj_result

    def __str__(self):
        """Get the status of the run.

        Returns:
            string: the status of the results.
        """
        return self._result['status']

    def __getitem__(self, i):
        return self._result['result'][i]

    def __len__(self):
        return len(self._result['result'])

    def __iadd__(self, other):
        """Append a Result object to current Result object.

        Arg:
            other (Result): a Result object to append.
        Returns:
            Result: The current object with appended results.
        Raises:
            QISKitError: if the Results cannot be combined.
        """
        # todo: reevaluate if moving equality to Backend themselves (part of
        # a bigger problem - backend instances will not persist between
        # sessions)
        this_backend = self._result.get('backend_name')
        other_backend = other._result.get('backend_name')
        if this_backend == other_backend:
            self._result['result'] += other._result['result']
            return self
        else:
            raise QISKitError('Result objects from different backends cannot be combined.')

    def __add__(self, other):
        """Combine Result objects.

        Arg:
            other (Result): a Result object to combine.
        Returns:
            Result: A new Result object consisting of combined objects.
        """
        ret = copy.deepcopy(self)
        ret += other
        return ret

    def _is_error(self):
        return self._result['status'] == 'ERROR'

    def get_status(self):
        """Return whole result status."""
        return self._result['status']

    def circuit_statuses(self):
        """Return statuses of all circuits

        Returns:
            list(str): List of status result strings.
        """
        return [circuit_result['status']
                for circuit_result in self._result['result']]

    def get_circuit_status(self, icircuit):
        """Return the status of circuit at index icircuit.

        Args:
            icircuit (int): index of circuit
        Returns:
            string: the status of the circuit.
        """
        return self._result['result'][icircuit]['status']

    def get_job_id(self):
        """Return the job id assigned by the api if this is a remote job.

        Returns:
            string: a string containing the job id.
        """
        return self._result['job_id']

    def get_ran_qasm(self, name):
        """Get the ran qasm for the named circuit and backend.

        Args:
            name (str): the name of the quantum circuit.

        Returns:
            string: A text version of the qasm file that has been run.
        Raises:
            QISKitError: if the circuit was not found.
        """
        try:
            for exp_result in self._result['result']:
                if exp_result.get('name') == name:
                    return exp_result['compiled_circuit_qasm']
        except KeyError:
            pass
        raise QISKitError('No  qasm for circuit "{0}"'.format(name))

    def get_data(self, circuit=None):
        """Get the data of circuit name.

        The data format will depend on the backend. For a real device it
        will be for the form::

            "counts": {'00000': XXXX, '00001': XXXX},
            "time"  : xx.xxxxxxxx

        for the qasm simulators of 1 shot::

            'statevector': array([ XXX,  ..., XXX]),
            'classical_state': 0

        for the qasm simulators of n shots::

            'counts': {'0000': XXXX, '1001': XXXX}

        for the unitary simulators::

            'unitary': np.array([[ XX + XXj
                                   ...
                                   XX + XX]
                                 ...
                                 [ XX + XXj
                                   ...
                                   XX + XXj]]

        Args:
            circuit (str or QuantumCircuit or None): reference to a quantum circuit
                If None and there is only one circuit available, returns
                that one.

        Returns:
            dict: A dictionary of data for the different backends.

        Raises:
            QISKitError: if there is no data for the circuit, or an unhandled
                error occurred while fetching the data.
            Exception: if a handled error occurred while fetching the data.
        """
        if self._is_error():
            exception = self._result['result']
            if isinstance(exception, BaseException):
                raise exception
            else:
                raise QISKitError(str(exception))
        if isinstance(circuit, QuantumCircuit):
            circuit = circuit.name

        if circuit is None:
            if len(self._result['result']) == 1:
                return self._result['result'][0]['data']
            else:
                raise QISKitError("You have to select a circuit when there is more than"
                                  "one available")

        if not isinstance(circuit, str):
            circuit = str(circuit)
        try:
            for circuit_result in self._result['result']:
                if circuit_result.get('name') == circuit:
                    return circuit_result['data']
        except (KeyError, TypeError):
            pass
        raise QISKitError('No data for circuit "{0}"'.format(circuit))

    def get_counts(self, circuit=None):
        """Get the histogram data of circuit name.

        The data from the a qasm circuit is dictionary of the format
        {'00000': XXXX, '00001': XXXXX}.

        Args:
            circuit (str or QuantumCircuit or None): reference to a quantum circuit
                If None and there is only one circuit available, returns
                that one.

        Returns:
            Dictionary: Counts {'00000': XXXX, '00001': XXXXX}.

        Raises:
            QISKitError: if there are no counts for the circuit.
        """
        try:
            return self.get_data(circuit)['counts']
        except KeyError:
            raise QISKitError('No counts for circuit "{0}"'.format(circuit))

    def get_statevector(self, circuit=None):
        """Get the final statevector of circuit name.

        The data is a list of complex numbers
        [1.+0.j, 0.+0.j].

        Args:
            circuit (str or QuantumCircuit or None): reference to a quantum circuit
                If None and there is only one circuit available, returns
                that one.

        Returns:
            list[complex]: list of 2^n_qubits complex amplitudes.

        Raises:
            QISKitError: if there is no statevector for the circuit.
        """
        try:
            return self.get_data(circuit)['statevector']
        except KeyError:
            raise QISKitError('No statevector for circuit "{0}"'.format(circuit))

    def get_unitary(self, circuit=None):
        """Get the final unitary of circuit name.

        The data is a matrix of complex numbers
        [[1.+0.j, 0.+0.j], .. ].

        Args:
            circuit (str or QuantumCircuit or None): reference to a quantum circuit
                If None and there is only one circuit available, returns
                that one.

        Returns:
            list[list[complex]]: list of 2^n_qubits x 2^n_qubits complex amplitudes.

        Raises:
            QISKitError: if there is no unitary for the circuit.
        """
        try:
            return self.get_data(circuit)['unitary']
        except KeyError:
            raise QISKitError('No unitary for circuit "{0}"'.format(circuit))

    def get_snapshots(self, circuit=None):
        """Get snapshots recorded during the run.

        The data is a dictionary:
        where keys are requested snapshot slots.
        and values are a dictionary of the snapshots themselves.

        Args:
            circuit (str or QuantumCircuit or None): reference to a quantum circuit
                If None and there is only one circuit available, returns
                that one.

        Returns:
            dict[slot: dict[str: array]]: list of 2^n_qubits complex amplitudes.

        Raises:
            QISKitError: if there are no snapshots for the circuit.
        """
        try:
            return self.get_data(circuit)['snapshots']
        except KeyError:
            raise QISKitError('No snapshots for circuit "{0}"'.format(circuit))

    def get_snapshot(self, slot=None, circuit=None):
        """Get snapshot at a specific slot.

        Args:
            slot (str): snapshot slot to retrieve. If None and there is only one
                slot, return that one.
            circuit (str or QuantumCircuit or None): reference to a quantum circuit
                If None and there is only one circuit available, returns
                that one.

        Returns:
            dict[slot: dict[str: array]]: list of 2^n_qubits complex amplitudes.

        Raises:
            QISKitError: if there is no snapshot at all, or in this slot
        """
        try:
            snapshots_dict = self.get_snapshots(circuit)

            if slot is None:
                slots = list(snapshots_dict.keys())
                if len(slots) == 1:
                    slot = slots[0]
                else:
                    raise QISKitError("You have to select a slot when there"
                                      "is more than one available")
            snapshot_dict = snapshots_dict[slot]

            snapshot_types = list(snapshot_dict.keys())
            if len(snapshot_types) == 1:
                snapshot_list = snapshot_dict[snapshot_types[0]]
                if len(snapshot_list) == 1:
                    return snapshot_list[0]
                else:
                    return snapshot_list
            else:
                return snapshot_dict
        except KeyError:
            raise QISKitError('No snapshot at slot {0} for '
                              'circuit "{1}"'.format(slot, circuit))

    def get_names(self):
        """Get the circuit names of the results.

        Returns:
            List: A list of circuit names.
        """
        return [c.get('name') for c in self._result['result']]

    def average_data(self, name, observable):
        """Compute the mean value of an diagonal observable.

        Takes in an observable in dictionary format and then
        calculates the sum_i value(i) P(i) where value(i) is the value of
        the observable for state i.

        Args:
            name (str): the name of the quantum circuit
            observable (dict): The observable to be averaged over. As an example
            ZZ on qubits equals {"00": 1, "11": 1, "01": -1, "10": -1}

        Returns:
            Double: Average of the observable
        """
        counts = self.get_counts(name)
        temp = 0
        tot = sum(counts.values())
        for key in counts:
            if key in observable:
                temp += counts[key] * observable[key] / tot
        return temp

    def get_qubitpol_vs_xval(self, nqubits, xvals_dict=None):
        """Compute the polarization of each qubit for all circuits and pull out each circuits
        xval into an array. Assumes that each circuit has the same number of qubits and that
        all qubits are measured.

        Args:
            nqubits (int): number of qubits
            xvals_dict (dict): xvals for each circuit {'circuitname1': xval1,...}. If this
            is none then the xvals list is just left as an array of zeros

        Returns:
            qubit_pol: mxn double array where m is the number of circuit, n the number of qubits
            xvals: mx1 array of the circuit xvals
        """
        ncircuits = len(self._result['result'])
        # Is this the best way to get the number of qubits?
        qubitpol = numpy.zeros([ncircuits, nqubits], dtype=float)
        xvals = numpy.zeros([ncircuits], dtype=float)

        # build Z operators for each qubit
        z_dicts = []
        for qubit_ind in range(nqubits):
            z_dicts.append(dict())
            for qubit_state in range(2**nqubits):
                new_key = ("{0:0"+"{:d}".format(nqubits) + "b}").format(qubit_state)
                z_dicts[-1][new_key] = -1
                if new_key[nqubits-qubit_ind-1] == '1':
                    z_dicts[-1][new_key] = 1

        # go through each circuit and for eqch qubit and apply the operators using "average_data"
        for circuit_ind in range(ncircuits):
            circuit_name = self._result['result'][circuit_ind]['name']
            if xvals_dict:
                xvals[circuit_ind] = xvals_dict[circuit_name]
            for qubit_ind in range(nqubits):
                qubitpol[circuit_ind, qubit_ind] = self.average_data(
                    circuit_name, z_dicts[qubit_ind])

        return qubitpol, xvals



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Node for an OPENQASM if statement.
"""
from ._node import Node


class If(Node):
    """Node for an OPENQASM if statement.

    children[0] is an id node.
    children[1] is an integer node.
    children[2] is quantum operation node, including U, CX, custom_unitary,
    measure, reset, (and BUG: barrier, if).
    """

    def __init__(self, children):
        """Create the if node."""
        Node.__init__(self, 'if', children, None)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        return "if(" + self.children[0].qasm(prec) + "==" \
               + str(self.children[1].value) + ") " + \
               self.children[2].qasm(prec)



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Backends provided by IBM Quantum Experience."""

from .ibmqprovider import IBMQProvider
from .ibmqbackend import IBMQBackend
from .ibmqjob import IBMQJob



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""
Node for an OPENQASM external function.
"""
import sympy
from ._node import Node
from ._nodeexception import NodeException


class External(Node):
    """Node for an OPENQASM external function.

    children[0] is an id node with the name of the function.
    children[1] is an expression node.
    """

    def __init__(self, children):
        """Create the external node."""
        Node.__init__(self, 'external', children, None)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        return self.children[0].qasm(prec) + "(" + \
            self.children[1].qasm(prec) + ")"

    def latex(self, prec=15, nested_scope=None):
        """Return the corresponding math mode latex string."""
        # pylint: disable=unused-argument
        # TODO prec ignored
        return sympy.latex(self.sym(nested_scope))

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        op = self.children[0].name
        expr = self.children[1]
        dispatch = {
            'sin': sympy.sin,
            'cos': sympy.cos,
            'tan': sympy.tan,
            'asin': sympy.asin,
            'acos': sympy.acos,
            'atan': sympy.atan,
            'exp': sympy.exp,
            'ln': sympy.log,
            'sqrt': sympy.sqrt
        }
        if op in dispatch:
            arg = expr.real(nested_scope)
            return dispatch[op](arg)
        else:
            raise NodeException("internal error: undefined external")

    def sym(self, nested_scope=None):
        """Return the corresponding symbolic expression."""
        op = self.children[0].name
        expr = self.children[1]
        dispatch = {
            'sin': sympy.sin,
            'cos': sympy.cos,
            'tan': sympy.tan,
            'asin': sympy.asin,
            'acos': sympy.acos,
            'atan': sympy.atan,
            'exp': sympy.exp,
            'ln': sympy.log,
            'sqrt': sympy.sqrt
        }
        if op in dispatch:
            arg = expr.sym(nested_scope)
            return dispatch[op](arg)
        else:
            raise NodeException("internal error: undefined external")



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""
Node for an OPENQASM qreg statement.
"""
from ._node import Node


class Qreg(Node):
    """Node for an OPENQASM qreg statement.

    children[0] is an indexedid node.
    """

    def __init__(self, children):
        """Create the qreg node."""
        Node.__init__(self, 'qreg', children, None)
        # This is the indexed id, the full "id[n]" object
        self.id = children[0]
        # Name of the qreg
        self.name = self.id.name
        # Source line number
        self.line = self.id.line
        # Source file name
        self.file = self.id.file
        # Size of the register
        self.index = self.id.index

    def to_string(self, indent):
        """Print the node data, with indent."""
        ind = indent * ' '
        print(ind, 'qreg')
        self.children[0].to_string(indent + 3)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        return "qreg " + self.id.qasm(prec) + ";"



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""
Node for an OPENQASM indexed id.
"""
from ._node import Node


class IndexedId(Node):
    """Node for an OPENQASM indexed id.

    children[0] is an id node.
    children[1] is an Int node.
    """

    def __init__(self, children):
        """Create the indexed id node."""
        Node.__init__(self, 'indexed_id', children, None)
        self.id = children[0]
        self.name = self.id.name
        self.line = self.id.line
        self.file = self.id.file
        self.index = children[1].value

    def to_string(self, indent):
        """Print with indent."""
        ind = indent * ' '
        print(ind, 'indexed_id', self.name, self.index)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        # pylint: disable=unused-argument
        return self.name + "[%d]" % self.index



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""This module implements the job class used for LocalBackend objects."""

from concurrent import futures
import logging
import sys

from qiskit.backends import BaseJob
from qiskit.backends import JobStatus
from qiskit import QISKitError

logger = logging.getLogger(__name__)


class LocalJob(BaseJob):
    """Local QISKit SDK Job class.

    Attributes:
        _executor (futures.Executor): executor to handle asynchronous jobs
    """

    if sys.platform in ['darwin', 'win32']:
        _executor = futures.ThreadPoolExecutor()
    else:
        _executor = futures.ProcessPoolExecutor()

    def __init__(self, fn, qobj):
        super().__init__()
        self._qobj = qobj
        self._backend_name = qobj['config']['backend_name']
        self._future = self._executor.submit(fn, qobj)

    def result(self, timeout=None):
        # pylint: disable=arguments-differ
        """
        Get job result. The behavior is the same as the underlying
        concurrent Future objects,

        https://docs.python.org/3/library/concurrent.futures.html#future-objects

        Args:
            timeout (float): number of seconds to wait for results.

        Returns:
            Result: Result object

        Raises:
            concurrent.futures.TimeoutError: if timeout occured.
            concurrent.futures.CancelledError: if job cancelled before completed.
        """
        return self._future.result(timeout=timeout)

    def cancel(self):
        return self._future.cancel()

    @property
    def status(self):
        _status_msg = None
        # order is important here
        if self.running:
            _status = JobStatus.RUNNING
        elif not self.done:
            _status = JobStatus.QUEUED
        elif self.cancelled:
            _status = JobStatus.CANCELLED
        elif self.done:
            _status = JobStatus.DONE
        elif self.exception:
            _status = JobStatus.ERROR
            _status_msg = str(self.exception)
        else:
            raise LocalJobError('Unexpected behavior of {0}'.format(
                self.__class__.__name__))
        return {'status': _status,
                'status_msg': _status_msg}

    @property
    def running(self):
        return self._future.running()

    @property
    def done(self):
        """
        Returns True if job successfully finished running.

        Note behavior is slightly different than Future objects which would
        also return true if successfully cancelled.
        """
        return self._future.done() and not self._future.cancelled()

    @property
    def cancelled(self):
        return self._future.cancelled()

    @property
    def exception(self):
        """
        Return Exception object if exception occured else None.

        Returns:
            Exception: exception raised by attempting to run job.
        """
        return self._future.exception(timeout=0)

    @property
    def backend_name(self):
        """
        Return backend name used for this job
        """
        return self._backend_name


class LocalJobError(QISKitError):
    """class for Local Job errors"""
    pass



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Standard extension's OPENQASM header and definition update.
"""
import sympy
from qiskit import QuantumCircuit
from qiskit.qasm import _node as node


if not hasattr(QuantumCircuit, '_extension_standard'):
    QuantumCircuit._extension_standard = True
    QuantumCircuit.header = QuantumCircuit.header + "\n" \
        + "include \"qelib1.inc\";"

    # 3-parameter 2-pulse single qubit gate
    QuantumCircuit.definitions["u3"] = {
        "print": False,
        "opaque": False,
        "n_args": 3,
        "n_bits": 1,
        "args": ["theta", "phi", "lambda"],
        "bits": ["q"],
        # gate u3(theta,phi,lambda) q { U(theta,phi,lambda) q; }
        "body": node.GateBody([
            node.UniversalUnitary([
                node.ExpressionList([
                    node.Id("theta", 0, ""),
                    node.Id("phi", 0, ""),
                    node.Id("lambda", 0, "")
                ]),
                node.Id("q", 0, "")
            ])
        ])
    }

    # 2-parameter 1-pulse single qubit gate
    QuantumCircuit.definitions["u2"] = {
        "print": False,
        "opaque": False,
        "n_args": 2,
        "n_bits": 1,
        "args": ["phi", "lambda"],
        "bits": ["q"],
        # gate u2(phi,lambda) q { U(pi/2,phi,lambda) q; }
        "body": node.GateBody([
            node.UniversalUnitary([
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Real(sympy.pi),
                        node.Int(2)
                    ]),
                    node.Id("phi", 0, ""),
                    node.Id("lambda", 0, "")
                ]),
                node.Id("q", 0, "")
            ])
        ])
    }

    # 1-parameter 0-pulse single qubit gate
    QuantumCircuit.definitions["u1"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 1,
        "args": ["lambda"],
        "bits": ["q"],
        # gate u1(lambda) q { U(0,0,lambda) q; }
        "body": node.GateBody([
            node.UniversalUnitary([
                node.ExpressionList([
                    node.Int(0),
                    node.Int(0),
                    node.Id("lambda", 0, "")
                ]),
                node.Id("q", 0, "")
            ])
        ])
    }

    # controlled-NOT
    QuantumCircuit.definitions["cx"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 2,
        "args": [],
        "bits": ["c", "t"],
        # gate cx c,t { CX c,t; }
        "body": node.GateBody([
            node.Cnot([
                node.Id("c", 0, ""),
                node.Id("t", 0, "")
            ])
        ])
    }

    # idle gate (identity)
    QuantumCircuit.definitions["id"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate id a { U(0,0,0) a; }
        "body": node.GateBody([
            node.UniversalUnitary([
                node.ExpressionList([
                    node.Int(0),
                    node.Int(0),
                    node.Int(0)
                ]),
                node.Id("a", 0, "")
            ])
        ])
    }

    # idle gate (identity) with length gamma*sqglen
    QuantumCircuit.definitions["u0"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 1,
        "args": ["gamma"],
        "bits": ["q"],
        # gate u0(gamma) q { U(0,0,0) q; }
        "body": node.GateBody([
            node.UniversalUnitary([
                node.ExpressionList([
                    node.Int(0),
                    node.Int(0),
                    node.Int(0)
                ]),
                node.Id("q", 0, "")
            ])
        ])
    }

    # Pauli gate: bit-flip
    QuantumCircuit.definitions["x"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate x a { u3(pi,0,pi) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u3", 0, ""),
                node.ExpressionList([
                    node.Real(sympy.pi),
                    node.Int(0),
                    node.Real(sympy.pi)
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Pauli gate: bit and phase flip
    QuantumCircuit.definitions["y"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate y a { u3(pi,pi/2,pi/2) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u3", 0, ""),
                node.ExpressionList([
                    node.Real(sympy.pi),
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Real(sympy.pi),
                        node.Int(2)
                    ]),
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Real(sympy.pi),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Pauli gate: phase flip
    QuantumCircuit.definitions["z"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate z a { u1(pi) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.Real(sympy.pi)
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Clifford gate: Hadamard
    QuantumCircuit.definitions["h"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate h a { u2(0,pi) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u2", 0, ""),
                node.ExpressionList([
                    node.Int(0),
                    node.Real(sympy.pi)
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Clifford gate: sqrt(Z) phase gate
    QuantumCircuit.definitions["s"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate s a { u1(pi/2) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Real(sympy.pi),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Clifford gate: conjugate of sqrt(Z)
    QuantumCircuit.definitions["sdg"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate sdg a { u1(-pi/2) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.Real(sympy.pi)
                        ]),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # C3 gate: sqrt(S) phase gate
    QuantumCircuit.definitions["t"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate t a { u1(pi/4) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Real(sympy.pi),
                        node.Int(4)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # C3 gate: conjugate of sqrt(S)
    QuantumCircuit.definitions["tdg"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 1,
        "args": [],
        "bits": ["a"],
        # gate tdg a { u1(-pi/4) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.Real(sympy.pi)
                        ]),
                        node.Int(4)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Rotation around X-axis
    QuantumCircuit.definitions["rx"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 1,
        "args": ["theta"],
        "bits": ["a"],
        # gate rx(theta) a { u3(theta, -pi/2,pi/2) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u3", 0, ""),
                node.ExpressionList([
                    node.Id("theta", 0, ""),
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.Real(sympy.pi)
                        ]),
                        node.Int(2)
                    ]),
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Real(sympy.pi),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Rotation around Y-axis
    QuantumCircuit.definitions["ry"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 1,
        "args": ["theta"],
        "bits": ["a"],
        # gate ry(theta) a { u3(theta,0,0) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u3", 0, ""),
                node.ExpressionList([
                    node.Id("theta", 0, ""),
                    node.Int(0),
                    node.Int(0)
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # Rotation around Z-axis
    QuantumCircuit.definitions["rz"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 1,
        "args": ["phi"],
        "bits": ["a"],
        # gate rz(phi) a { u1(phi) a; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.Id("phi", 0, "")
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # controlled-Phase
    QuantumCircuit.definitions["cz"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 2,
        "args": [],
        "bits": ["a", "b"],
        # gate cz a,b { h b; cx a,b; h b; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # controlled-Y
    QuantumCircuit.definitions["cy"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 2,
        "args": [],
        "bits": ["a", "b"],
        # gate cy a,b { sdg b; cx a,b; s b; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("sdg", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("s", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # swap
    QuantumCircuit.definitions["swap"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 2,
        "args": [],
        "bits": ["a", "b"],
        # gate swap a,b { cx a,b; cx b,a; cx a,b; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, ""),
                    node.Id("a", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # controlled-H
    QuantumCircuit.definitions["ch"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 2,
        "args": [],
        "bits": ["a", "b"],
        # gate ch a,b {
        # h b; sdg b;
        # cx a,b;
        # h b; t b;
        # cx a,b;
        # t b; h b; s b; x b; s a;
        # }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("sdg", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("t", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("t", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("s", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("x", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("s", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ])
        ])
    }

    # C3 gate: Toffoli
    QuantumCircuit.definitions["ccx"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 3,
        "args": [],
        "bits": ["a", "b", "c"],
        # gate ccx a,b,c
        # {
        #   h c;
        #   cx b,c; tdg c;
        #   cx a,c; t c;
        #   cx b,c; tdg c;
        #   cx a,c; t b; t c; h c;
        #   cx a,b; t a; tdg b;
        #   cx a,b;
        # }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, ""),
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("tdg", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("t", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, ""),
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("tdg", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("t", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("t", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("h", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("t", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("tdg", 0, ""),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # cswap (Fredkin)
    QuantumCircuit.definitions["cswap"] = {
        "print": False,
        "opaque": False,
        "n_args": 0,
        "n_bits": 3,
        "args": [],
        "bits": ["a", "b", "c"],
        # gate cswap a,b,c
        # {
        #  cx c,b;
        #  ccx a,b,c;
        #  cx c,b;
        # }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("ccx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, ""),
                    node.Id("c", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, ""),
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # controlled rz rotation
    QuantumCircuit.definitions["crz"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 2,
        "args": ["lambda"],
        "bits": ["a", "b"],
        # gate crz(lambda) a,b
        # {
        # u1(lambda/2) b;
        # cx a,b;
        # u1(-lambda/2) b;
        # cx a,b;
        # }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Id("lambda", 0, ""),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.Id("lambda", 0, "")
                        ]),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # controlled phase rotation
    QuantumCircuit.definitions["cu1"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 2,
        "args": ["lambda"],
        "bits": ["a", "b"],
        # gate cu1(lambda) a,b
        # {
        # u1(lambda/2) a;
        # cx a,b;
        # u1(-lambda/2) b;
        # cx a,b;
        # u1(lambda/2) b;
        # }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Id("lambda", 0, ""),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("a", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.Id("lambda", 0, "")
                        ]),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Id("lambda", 0, ""),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }

    # controlled-U
    QuantumCircuit.definitions["cu3"] = {
        "print": False,
        "opaque": False,
        "n_args": 3,
        "n_bits": 2,
        "args": ["theta", "phi", "lambda"],
        "bits": ["c", "t"],
        # gate cu3(theta,phi,lambda) c, t
        # {
        #  u1((lambda-phi)/2) t;
        #  cx c,t;
        #  u3(-theta/2,0,-(phi+lambda)/2) t;
        #  cx c,t;
        #  u3(theta/2,phi,0) t;
        # }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.BinaryOp([
                            node.BinaryOperator('-'),
                            node.Id("lambda", 0, ""),
                            node.Id("phi", 0, "")
                        ]),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("t", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, ""),
                    node.Id("t", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("u3", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.Id("theta", 0, "")
                        ]),
                        node.Int(2)
                    ]),
                    node.Int(0),
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Prefix([
                            node.UnaryOperator('-'),
                            node.BinaryOp([
                                node.BinaryOperator('+'),
                                node.Id("phi", 0, ""),
                                node.Id("lambda", 0, "")
                            ]),
                        ]),
                        node.Int(2)
                    ])
                ]),
                node.PrimaryList([
                    node.Id("t", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("c", 0, ""),
                    node.Id("t", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("u3", 0, ""),
                node.ExpressionList([
                    node.BinaryOp([
                        node.BinaryOperator('/'),
                        node.Id("theta", 0, ""),
                        node.Int(2)
                    ]),
                    node.Id("phi", 0, ""),
                    node.Int(0)
                ]),
                node.PrimaryList([
                    node.Id("t", 0, "")
                ])
            ])
        ])
    }

    # Two-qubit ZZ-rotation by angle theta
    QuantumCircuit.definitions["rzz"] = {
        "print": False,
        "opaque": False,
        "n_args": 1,
        "n_bits": 2,
        "args": ["theta"],
        "bits": ["a", "b"],
        # gate rzz(theta) a, b { cx a, b; u1(theta) b; cx a, b; }
        "body": node.GateBody([
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("u1", 0, ""),
                node.ExpressionList([
                    node.Id("theta", 0, "")
                ]),
                node.PrimaryList([
                    node.Id("b", 0, "")
                ])
            ]),
            node.CustomUnitary([
                node.Id("cx", 0, ""),
                node.PrimaryList([
                    node.Id("a", 0, ""),
                    node.Id("b", 0, "")
                ])
            ])
        ])
    }



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
OPENQASM circuit object.
"""
from ._qasmerror import QasmError
from ._qasmparser import QasmParser


class Qasm(object):
    """OPENQASM circuit object."""

    def __init__(self, filename=None, data=None):
        """Create an OPENQASM circuit object."""
        if filename is None and data is None:
            raise QasmError("Missing input file and/or data")
        if filename is not None and data is not None:
            raise QasmError("File and data must not both be specified"
                            "initializing qasm")
        self._filename = filename
        self._data = data

    def get_filename(self):
        """Return the filename."""
        return self._filename

    def get_tokens(self):
        """Returns a generator of the tokens."""
        if self._filename:
            with open(self._filename) as ifile:
                self._data = ifile.read()

        with QasmParser(self._filename) as qasm_p:
            return qasm_p.get_tokens()

    def parse(self):
        """Parse the data."""
        if self._filename:
            with open(self._filename) as ifile:
                self._data = ifile.read()

        with QasmParser(self._filename) as qasm_p:
            qasm_p.parse_debug(False)
            return qasm_p.parse(self._data)



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Simulator command to save the quantum state.
"""
from qiskit import Instruction
from qiskit import QuantumCircuit
from qiskit import CompositeGate
from qiskit import QuantumRegister
from qiskit.extensions._extensionerror import ExtensionError
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class Save(Instruction):
    """Simulator save instruction."""

    def __init__(self, slot, qubits, circ):
        """Create save save instruction."""
        super().__init__("save", [slot], list(qubits), circ)

    def inverse(self):
        """Special case. Return self."""
        return self

    def qasm(self):
        """Return OPENQASM string."""
        string = "save(%d) " % self.param[0]
        for j in range(len(self.arg)):
            if len(self.arg[j]) == 1:
                string += "%s" % self.arg[j].name
            else:
                string += "%s[%d]" % (self.arg[j][0].name, self.arg[j][1])
            if j != len(self.arg) - 1:
                string += ","
        string += ";"
        return string

    def reapply(self, circ):
        """Reapply this instruction to corresponding qubits in circ."""
        self._modifiers(circ.save(self.param[0]))


def save(self, slot):
    """Save the internal simulator representation (statevector, probability,
    density matrix, clifford table).
    Works on all qubits, and prevents reordering (like barrier).

    Args:
        slot (int): a slot to save into

    Returns:
        QuantumCircuit: with attached command

    Raises:
        ExtensionError: malformed command
    """
    tuples = []
    if isinstance(self, QuantumCircuit):
        for register in self.regs.values():
            if isinstance(register, QuantumRegister):
                tuples.append(register)
    if not tuples:
        raise ExtensionError("no qubits for save")
    if slot is None:
        raise ExtensionError("no save slot passed")
    qubits = []
    for tuple_element in tuples:
        if isinstance(tuple_element, QuantumRegister):
            for j in range(tuple_element.size):
                self._check_qubit((tuple_element, j))
                qubits.append((tuple_element, j))
        else:
            self._check_qubit(tuple_element)
            qubits.append(tuple_element)
    self._check_dups(qubits)
    return self._attach(Save(slot, qubits, self))


# Add to QuantumCircuit and CompositeGate classes
QuantumCircuit.save = save
CompositeGate.save = save



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
A module for using quaternions.
"""
import math
import numpy as np
import scipy.linalg as la


class Quaternion:
    """A class representing a Quaternion.
    """

    def __init__(self, data):
        self.data = np.asarray(data, dtype=float)

    def __call__(self, idx):
        return self.data[idx]

    def __repr__(self):
        return np.array_str(self.data)

    def __str__(self):
        return np.array_str(self.data)

    def __mul__(self, r):  # pylint: disable=C0103
        if isinstance(r, Quaternion):
            q = self  # pylint: disable=C0103
            out_data = np.zeros(4, dtype=float)
            out_data[0] = r(0)*q(0) - r(1)*q(1) - r(2)*q(2) - r(3)*q(3)
            out_data[1] = r(0)*q(1) + r(1)*q(0) - r(2)*q(3) + r(3)*q(2)
            out_data[2] = r(0)*q(2) + r(1)*q(3) + r(2)*q(0) - r(3)*q(1)
            out_data[3] = r(0)*q(3) - r(1)*q(2) + r(2)*q(1) + r(3)*q(0)
            return Quaternion(out_data)
        else:
            raise Exception('Multiplication by other not supported.')

    def norm(self):
        """ Norm of quaternion.
        """
        return la.norm(self.data)

    def normalize(self, inplace=False):
        """Normalizes a Quaternion to unit length
        so that it represents a valid roation.

        Args:
            inplace (bool): Do an inplace normalization.

        Returns:
            Quaternion: Normalized quaternion.
        """
        if inplace:
            nrm = self.norm()
            self.data /= nrm
            return None
        nrm = self.norm()
        data_copy = np.array(self.data, copy=True)
        data_copy /= nrm
        return Quaternion(data_copy)

    def to_matrix(self):
        """Converts a unit-length quaternion to a rotation matrix.

        Returns:
            ndarray: Rotation matrix.
        """
        w, x, y, z = self.normalize().data  # pylint: disable=C0103
        mat = np.array([
            [1-2*y**2-2*z**2, 2*x*y-2*z*w, 2*x*z+2*y*w],
            [2*x*y+2*z*w, 1-2*x**2-2*z**2, 2*y*z-2*x*w],
            [2*x*z-2*y*w, 2*y*z+2*x*w, 1-2*x**2-2*y**2]
        ], dtype=float)
        return mat

    def to_zyz(self):
        """Converts a unit-length quaternion to a sequence
        of ZYZ Euler angles.

        Returns:
            ndarray: Array of Euler angles.
        """
        mat = self.to_matrix()
        euler = np.zeros(3, dtype=float)
        if mat[2, 2] < 1:
            if mat[2, 2] > -1:
                euler[0] = math.atan2(mat[1, 2], mat[0, 2])
                euler[1] = math.acos(mat[2, 2])
                euler[2] = math.atan2(mat[2, 1], -mat[2, 0])
            else:
                euler[0] = -math.atan2(mat[1, 0], mat[1, 1])
                euler[1] = np.pi
        else:
            euler[0] = math.atan2(mat[1, 0], mat[1, 1])
        return euler


def quaternion_from_axis_rotation(angle, axis):
    """Return quaternion for rotation about given axis.

    Args:
        angle (float): Angle in radians.
        axis (str): Axis for rotation

    Returns:
        Quaternion: Quaternion for axis rotation.

    Raises:
        ValueError: Invalid input axis.
    """
    out = np.zeros(4, dtype=float)
    if axis == 'x':
        out[1] = 1
    elif axis == 'y':
        out[2] = 1
    elif axis == 'z':
        out[3] = 1
    else:
        raise ValueError('Invalid axis input.')
    out *= math.sin(angle/2.0)
    out[0] = math.cos(angle/2.0)
    return Quaternion(out)


def quaternion_from_euler(angles, order='yzy'):
    """Generate a quaternion from a set of Euler angles.

    Args:
        angles (array_like): Array of Euler angles.
        order (str): Order of Euler rotations.  'yzy' is default.

    Returns:
        Quaternion: Quaternion representation of Euler rotation.
    """
    angles = np.asarray(angles, dtype=float)
    quat = quaternion_from_axis_rotation(angles[0], order[0])\
        * (quaternion_from_axis_rotation(angles[1], order[1])
           * quaternion_from_axis_rotation(angles[2], order[2]))
    quat.normalize(inplace=True)
    return quat


def _rotm(angle, axis):
    """Generates a rotation matrix for a given angle and axis.

    Args:
        angle (float): Roation angle in radians.
        axis (str): Axis for rotation: 'x', 'y', 'z'

    Returns:
        ndarray: Rotation matrix.

    Raises:
        ValueError: Invalid input axis.
    """
    direction = np.zeros(3, dtype=float)
    if axis == 'x':
        direction[0] = 1
    elif axis == 'y':
        direction[1] = 1
    elif axis == 'z':
        direction[2] = 1
    else:
        raise ValueError('Invalid axis.')
    direction = np.asarray(direction, dtype=float)
    sin_angle = math.sin(angle)
    cos_angle = math.cos(angle)
    rot = np.diag([cos_angle, cos_angle, cos_angle])
    rot += np.outer(direction, direction) * (1.0 - cos_angle)
    direction *= sin_angle
    rot += np.array([
        [0, -direction[2], direction[1]],
        [direction[2], 0, -direction[0]],
        [-direction[1], direction[0], 0]
    ])
    return rot



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Interface to C++ quantum circuit simulator with realistic noise.
"""

import json
import logging
import os
import subprocess
from subprocess import PIPE
import platform
import warnings

import numpy as np

from qiskit._result import Result
from qiskit.backends import BaseBackend
from qiskit.backends.local.localjob import LocalJob

logger = logging.getLogger(__name__)

EXTENSION = '.exe' if platform.system() == 'Windows' else ''

# Add path to compiled qasm simulator
DEFAULT_SIMULATOR_PATHS = [
    # This is the path where Makefile creates the simulator by default
    os.path.abspath(os.path.join(os.path.dirname(__file__),
                                 '../../../out/src/qasm-simulator-cpp/qasm_simulator_cpp'
                                 + EXTENSION)),
    # This is the path where PIP installs the simulator
    os.path.abspath(os.path.join(os.path.dirname(__file__),
                                 'qasm_simulator_cpp' + EXTENSION)),
]


class QasmSimulatorCpp(BaseBackend):
    """C++ quantum circuit simulator with realistic noise"""

    DEFAULT_CONFIGURATION = {
        'name': 'local_qasm_simulator_cpp',
        'url': 'https://github.com/QISKit/qiskit-terra/src/qasm-simulator-cpp',
        'simulator': True,
        'local': True,
        'description': 'A C++ realistic noise simulator for qobj files',
        'coupling_map': 'all-to-all',
        "basis_gates": 'u0,u1,u2,u3,cx,cz,id,x,y,z,h,s,sdg,t,tdg,rzz,' +
                       'snapshot,wait,noise,save,load'
    }

    def __init__(self, configuration=None):
        super().__init__(configuration or self.DEFAULT_CONFIGURATION.copy())
        # Try to use the default executable if not specified.
        if self._configuration.get('exe'):
            paths = [self._configuration.get('exe')]
        else:
            paths = DEFAULT_SIMULATOR_PATHS

        # Ensure that the executable is available.
        try:
            self._configuration['exe'] = next(
                path for path in paths if (os.path.exists(path) and
                                           os.path.getsize(path) > 100))
        except StopIteration:
            raise FileNotFoundError('Simulator executable not found (using %s)' %
                                    self._configuration.get('exe', 'default locations'))

    def run(self, qobj):
        """Run a qobj on the backend."""
        return LocalJob(self._run_job, qobj)

    def _run_job(self, qobj):
        self._validate(qobj)
        result = run(qobj, self._configuration['exe'])
        return Result(result)

    def _validate(self, qobj):
        if qobj['config']['shots'] == 1:
            warnings.warn('The behavior of getting statevector from simulators '
                          'by setting shots=1 is deprecated and will be removed. '
                          'Use the local_statevector_simulator instead, or place '
                          'explicit snapshot instructions.',
                          DeprecationWarning)
        for circ in qobj['circuits']:
            if 'measure' not in [op['name'] for
                                 op in circ['compiled_circuit']['operations']]:
                logger.warning("no measurements in circuit '%s', "
                               "classical register will remain all zeros.", circ['name'])
        return


class CliffordSimulatorCpp(BaseBackend):
    """"C++ Clifford circuit simulator with realistic noise."""

    DEFAULT_CONFIGURATION = {
        'name': 'local_clifford_simulator_cpp',
        'url': 'https://github.com/QISKit/qiskit-terra/src/qasm-simulator-cpp',
        'simulator': True,
        'local': True,
        'description': 'A C++ Clifford simulator with approximate noise',
        'coupling_map': 'all-to-all',
        'basis_gates': 'cx,id,x,y,z,h,s,sdg,snapshot,wait,noise,save,load'
    }

    def __init__(self, configuration=None):
        super().__init__(configuration or self.DEFAULT_CONFIGURATION.copy())

        # Try to use the default executable if not specified.
        if self._configuration.get('exe'):
            paths = [self._configuration.get('exe')]
        else:
            paths = DEFAULT_SIMULATOR_PATHS

        # Ensure that the executable is available.
        try:
            self._configuration['exe'] = next(
                path for path in paths if (os.path.exists(path) and
                                           os.path.getsize(path) > 100))
        except StopIteration:
            raise FileNotFoundError('Simulator executable not found (using %s)' %
                                    self._configuration.get('exe', 'default locations'))

    def run(self, qobj):
        """Run a Qobj on the backend.

        Args:
            qobj (dict): job description

        Returns:
            LocalJob: derived from BaseJob
        """
        return LocalJob(self._run_job, qobj)

    def _run_job(self, qobj):
        self._validate()
        # set backend to Clifford simulator
        if 'config' in qobj:
            qobj['config']['simulator'] = 'clifford'
        else:
            qobj['config'] = {'simulator': 'clifford'}

        result = run(qobj, self._configuration['exe'])
        return Result(result)

    def _validate(self):
        return


class QASMSimulatorEncoder(json.JSONEncoder):
    """
    JSON encoder for NumPy arrays and complex numbers.

    This functions as the standard JSON Encoder but adds support
    for encoding:
        complex numbers z as lists [z.real, z.imag]
        ndarrays as nested lists.
    """

    # pylint: disable=method-hidden,arguments-differ
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, complex):
            return [obj.real, obj.imag]
        return json.JSONEncoder.default(self, obj)


class QASMSimulatorDecoder(json.JSONDecoder):
    """
    JSON decoder for the output from C++ qasm_simulator.

    This converts complex vectors and matrices into numpy arrays
    for the following keys.
    """
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    # pylint: disable=method-hidden
    def object_hook(self, obj):
        """Special decoding rules for simulator output."""

        for key in ['U_error', 'density_matrix']:
            # JSON is a complex matrix
            if key in obj and isinstance(obj[key], list):
                tmp = np.array(obj[key])
                obj[key] = tmp[::, ::, 0] + 1j * tmp[::, ::, 1]
        for key in ['statevector', 'inner_products']:
            # JSON is a list of complex vectors
            if key in obj:
                for j in range(len(obj[key])):
                    if isinstance(obj[key][j], list):
                        tmp = np.array(obj[key][j])
                        obj[key][j] = tmp[::, 0] + 1j * tmp[::, 1]
        return obj


def run(qobj, executable):
    """
    Run simulation on C++ simulator inside a subprocess.

    Args:
        qobj (dict): qobj dictionary defining the simulation to run
        executable (string): filename (with path) of the simulator executable
    Returns:
        dict: A dict of simulation results
    """

    # Open subprocess and execute external command
    try:
        with subprocess.Popen([executable, '-'],
                              stdin=PIPE, stdout=PIPE, stderr=PIPE) as proc:
            cin = json.dumps(qobj, cls=QASMSimulatorEncoder).encode()
            cout, cerr = proc.communicate(cin)
        if cerr:
            logger.error('ERROR: Simulator encountered a runtime error: %s',
                         cerr.decode())
        sim_output = cout.decode()
        return json.loads(sim_output, cls=QASMSimulatorDecoder)

    except FileNotFoundError:
        msg = "ERROR: Simulator exe not found at: %s" % executable
        logger.error(msg)
        return {"status": msg, "success": False}


def cx_error_matrix(cal_error, zz_error):
    """
    Return the coherent error matrix for CR error model of a CNOT gate.

    Args:
        cal_error (double): calibration error of rotation
        zz_error (double): ZZ interaction term error

    Returns:
        numpy.ndarray: A coherent error matrix U_error for the CNOT gate.

    Details:

    The ideal cross-resonsance (CR) gate corresponds to a 2-qubit rotation
        U_CR_ideal = exp(-1j * (pi/2) * XZ/2)
    where qubit-0 is the control, and qubit-1 is the target. This can be
    converted to a CNOT gate by single-qubit rotations
        U_CX = U_L * U_CR_ideal * U_R.

    The noisy rotation is implemented as
        U_CR_noise = exp(-1j * (pi/2 + cal_error) * (XZ + zz_error ZZ)/2)

    The retured error matrix is given by
        U_error = U_L * U_CR_noise * U_R * U_CX^dagger
    """
    # pylint: disable=invalid-name
    if cal_error == 0 and zz_error == 0:
        return np.eye(4)

    cx_ideal = np.array([[1, 0, 0, 0],
                         [0, 0, 0, 1],
                         [0, 0, 1, 0],
                         [0, 1, 0, 0]])
    b = np.sqrt(1.0 + zz_error * zz_error)
    a = b * (np.pi / 2.0 + cal_error) / 2.0
    sp = (1.0 + 1j * zz_error) * np.sin(a) / b
    sm = (1.0 - 1j * zz_error) * np.sin(a) / b
    c = np.cos(a)
    cx_noise = np.array([[c + sm, 0, -1j * (c - sm), 0],
                         [0, 1j * (c - sm), 0, c + sm],
                         [-1j * (c - sp), 0, c + sp, 0],
                         [0, c + sp, 0, 1j * (c - sp)]]) / np.sqrt(2)
    return cx_noise.dot(cx_ideal.conj().T)


def x90_error_matrix(cal_error, detuning_error):
    """
    Return the coherent error matrix for a X90 rotation gate.

    Args:
        cal_error (double): calibration error of rotation
        detuning_error (double): detuning amount for rotation axis error

    Returns:
        numpy.ndarray: A coherent error matrix U_error for the X90 gate.

    Details:

    The ideal X90 rotation is a pi/2 rotation about the X-axis:
        U_X90_ideal = exp(-1j (pi/2) X/2)
    The noisy rotation is implemented as
        U_X90_noise = exp(-1j (pi/2 + cal_error) (cos(d) X + sin(d) Y)/2)
    where d is the detuning_error.

    The retured error matrix is given by
        U_error = U_X90_noise * U_X90_ideal^dagger
    """
    # pylint: disable=invalid-name
    if cal_error == 0 and detuning_error == 0:
        return np.eye(2)
    else:
        x90_ideal = np.array([[1., -1.j], [-1.j, 1]]) / np.sqrt(2)
        c = np.cos(0.5 * cal_error)
        s = np.sin(0.5 * cal_error)
        gamma = np.exp(-1j * detuning_error)
        x90_noise = np.array([[c - s, -1j * (c + s) * gamma],
                              [-1j * (c + s) * np.conj(gamma), c - s]]) / np.sqrt(2)
    return x90_noise.dot(x90_ideal.conj().T)


def _generate_coherent_error_matrix(config):
    """
    Generate U_error matrix for CX and X90 gates.

    Args:
        config (dict): the config of a qobj circuit

    This parses the config for the following noise parameter keys and returns a
    coherent error matrix for simulation coherent noise.
        'CX' gate: 'calibration_error', 'zz_error'
        'X90' gate: 'calibration_error', 'detuning_error'
    """
    # pylint: disable=invalid-name
    if 'noise_params' in config:
        # Check for CR coherent error parameters
        if 'CX' in config['noise_params']:
            noise_cx = config['noise_params']['CX']
            cal_error = noise_cx.pop('calibration_error', 0)
            zz_error = noise_cx.pop('zz_error', 0)
            # Add to current coherent error matrix
            if not cal_error == 0 or not zz_error == 0:
                u_error = noise_cx.get('U_error', np.eye(4))
                u_error = u_error.dot(cx_error_matrix(cal_error, zz_error))
                config['noise_params']['CX']['U_error'] = u_error
        # Check for X90 coherent error parameters
        if 'X90' in config['noise_params']:
            noise_x90 = config['noise_params']['X90']
            cal_error = noise_x90.pop('calibration_error', 0)
            detuning_error = noise_x90.pop('detuning_error', 0)
            # Add to current coherent error matrix
            if not cal_error == 0 or not detuning_error == 0:
                u_error = noise_x90.get('U_error', np.eye(2))
                u_error = u_error.dot(x90_error_matrix(cal_error,
                                                       detuning_error))
                config['noise_params']['X90']['U_error'] = u_error



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""This module implements the base pass.
"""
from abc import ABC, abstractmethod


class BasePass(ABC):
    """Base class for transpiler passes."""

    @abstractmethod
    def run(self, dag):
        """Run a pass on the DAGCircuit.
        """
        return dag



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
two-qubit ZZ-rotation gate.
"""
from qiskit import CompositeGate
from qiskit import Gate
from qiskit import QuantumCircuit
from qiskit._instructionset import InstructionSet
from qiskit._quantumregister import QuantumRegister
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class RZZGate(Gate):
    """Two-qubit ZZ-rotation gate."""

    def __init__(self, theta, ctl, tgt, circ=None):
        """Create new rzz gate."""
        super().__init__("rzz", [theta], [ctl, tgt], circ)

    def qasm(self):
        """Return OPENQASM string."""
        ctl = self.arg[0]
        tgt = self.arg[1]
        theta = self.param[0]
        return self._qasmif("rzz(%s) %s[%d],%s[%d];" % (theta,
                                                        ctl[0].name, ctl[1],
                                                        tgt[0].name, tgt[1]))

    def inverse(self):
        """Invert this gate."""
        self.param[0] = -self.param[0]
        return self

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.rzz(self.param[0], self.arg[0], self.arg[1]))


def rzz(self, theta, ctl, tgt):
    """Apply RZZ to circuit."""
    if isinstance(ctl, QuantumRegister) and \
       isinstance(tgt, QuantumRegister) and len(ctl) == len(tgt):
        instructions = InstructionSet()
        for i in range(ctl.size):
            instructions.add(self.rzz(theta, (ctl, i), (tgt, i)))
        return instructions

    self._check_qubit(ctl)
    self._check_qubit(tgt)
    self._check_dups([ctl, tgt])
    return self._attach(RZZGate(theta, ctl, tgt, self))


# Add to QuantumCircuit and CompositeGate classes
QuantumCircuit.rzz = rzz
CompositeGate.rzz = rzz



# -*- coding: utf-8 -*-
# pylint: disable=redefined-builtin

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""This module implements a passmanager
"""


class PassManager():
    """PassManager class for the transpiler.

    A PassManager instance is responsible for launching the requisite
    analysis and transformation passes on the quantum circuit, and to
    do this correctly & efficiently
    (i.e. keep track of dependencies between passes)
    """
    def __init__(self):
        """Initialize an empty PassManager object
        (with no passes scheduled).
        """
        self._passes = []

    def add_pass(self, pass_):
        """Schedule a pass in the passmanager."""
        self._passes.append(pass_)

    def passes(self):
        """Return list of passes scheduled."""
        return self._passes



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Utils for transpiler."""

from ._passmanager import PassManager
from ._transpilererror import TranspilerError

# pylint: disable=redefined-builtin
from ._transpiler import compile, transpile



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Provider for remote IbmQ backends."""
from IBMQuantumExperience import IBMQuantumExperience

from qiskit._util import _camel_case_to_snake_case
from qiskit.backends.baseprovider import BaseProvider
from qiskit.backends.ibmq.ibmqbackend import IBMQBackend, ALIASED_BACKEND_NAMES_REVERSED


class IBMQProvider(BaseProvider):
    """Provider for remote IbmQ backends."""
    def __init__(self, token, url='https://quantumexperience.ng.bluemix.net/api',
                 hub=None, group=None, project=None, proxies=None, verify=True):
        super().__init__()

        # Get a connection to IBMQuantumExperience.
        self._api = self._authenticate(token, url,
                                       hub, group, project, proxies, verify)

        # Populate the list of remote backends.
        self.backends = self._discover_remote_backends()

        # authentication attributes, which uniquely identify the provider instance
        self._token = token
        self._url = url
        self._hub = hub
        self._group = group
        self._project = project
        self._proxies = proxies
        self._verify = verify

    def get_backend(self, name):
        return self.backends[name]

    def available_backends(self):
        """Get a list of available backends from the IBMQ provider.

        Returns:
            list[IBMQBackend]: a list of backend instances available
            from the IBMQ provider.
        """
        # pylint: disable=arguments-differ
        return list(self.backends.values())

    def grouped_backend_names(self):
        return {}

    def deprecated_backend_names(self):
        return {
            'ibmqx_qasm_simulator': 'ibmq_qasm_simulator',
            'ibmqx_hpc_qasm_simulator': 'ibmq_qasm_simulator',
            'real': 'ibmqx1'
            }

    def aliased_backend_names(self):
        return ALIASED_BACKEND_NAMES_REVERSED

    @classmethod
    def _authenticate(cls, token, url,
                      hub=None, group=None, project=None, proxies=None,
                      verify=True):
        """
        Authenticate against the IBMQuantumExperience API.

        Returns:
            IBMQuantumExperience.IBMQuantumExperience.IBMQuantumExperience:
                instance of the IBMQuantumExperience API.
        Raises:
            ConnectionError: if the authentication resulted in error.
        """
        try:
            config_dict = {
                'url': url,
            }
            # Only append hub/group/project if they are different than None.
            if all([hub, group, project]):
                config_dict.update({
                    'hub': hub,
                    'group': group,
                    'project': project
                })
            if proxies:
                config_dict['proxies'] = proxies
            return IBMQuantumExperience(token, config_dict, verify)
        except Exception as ex:
            root_exception = ex
            if 'License required' in str(ex):
                # For the 401 License required exception from the API, be
                # less verbose with the exceptions.
                root_exception = None
            raise ConnectionError("Couldn't connect to IBMQuantumExperience server: {0}"
                                  .format(ex)) from root_exception

    @classmethod
    def _parse_backend_configuration(cls, config):
        """
        Parse a backend configuration returned by IBMQuantumConfiguration.

        Args:
            config (dict): raw configuration as returned by
                IBMQuantumExperience.

        Returns:
            dict: parsed configuration.
        """
        edited_config = {
            'local': False
        }

        for key in config.keys():
            new_key = _camel_case_to_snake_case(key)
            if new_key not in ['id', 'serial_number', 'topology_id',
                               'status']:
                edited_config[new_key] = config[key]

        # FIXME (new_backend_names): a hack to show the new device display
        # names. Needs to be fixed in the API.
        if config.get('name') in ALIASED_BACKEND_NAMES_REVERSED.keys():
            # Use the new name of the backend for `backend.name` if it is in
            # the list of aliases.
            edited_config['name'] = ALIASED_BACKEND_NAMES_REVERSED[config['name']]

        return edited_config

    def _discover_remote_backends(self):
        """
        Return the remote backends available.

        Returns:
            dict[str:IBMQBackend]: a dict of the remote backend instances,
                keyed by backend name.
        """
        ret = {}
        configs_list = self._api.available_backends()
        for raw_config in configs_list:
            config = self._parse_backend_configuration(raw_config)
            ret[config['name']] = IBMQBackend(configuration=config, api=self._api)

        return ret

    def __eq__(self, other):
        try:
            equality = (self._token == other._token and self._url == other._url and
                        self._hub == other._hub and self._group == other._group and
                        self._project == other._project)
        except AttributeError:
            equality = False
        return equality



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Meta-provider that aggregates several providers."""

import logging
from itertools import combinations

from qiskit import QISKitError
from qiskit.backends.baseprovider import BaseProvider
from qiskit.backends.local.localprovider import LocalProvider

logger = logging.getLogger(__name__)


class DefaultQISKitProvider(BaseProvider):
    """
    Meta-provider that aggregates several providers.
    """
    def __init__(self):
        super().__init__()

        # List of providers.
        self.providers = [LocalProvider()]

    def get_backend(self, name):
        name = self.resolve_backend_name(name)
        for provider in self.providers:
            try:
                return provider.get_backend(name)
            except KeyError:
                pass
        raise KeyError(name)

    def available_backends(self, filters=None):
        """Get a list of available backends from all providers (after filtering).

        Note:
            If two or more providers share similar backend names, only the backends
            belonging to the 1st registered provider will be returned.

        Args:
            filters (dict or callable): filtering conditions.
                each will either pass through, or be filtered out.
                1) dict: {'criteria': value}
                    the criteria can be over backend's `configuration` or `status`
                    e.g. {'local': False, 'simulator': False, 'operational': True}
                2) callable: BaseBackend -> bool
                    e.g. lambda x: x.configuration['n_qubits'] > 5

        Returns:
            list[BaseBackend]: a list of backend instances available
                from all the providers.

        Raises:
            QISKitError: if passing filters that is neither dict nor callable
        """
        # pylint: disable=arguments-differ
        backends = []
        for provider in self.providers:
            backends.extend(provider.available_backends())

        if filters is not None:
            if isinstance(filters, dict):
                # exact match filter:
                # e.g. {'n_qubits': 5, 'operational': True}
                for key, value in filters.items():
                    backends = [instance for instance in backends
                                if instance.configuration.get(key) == value
                                or instance.status.get(key) == value]
            elif callable(filters):
                # acceptor filter: accept or reject a specific backend
                # e.g. lambda x: x.configuration['n_qubits'] > 5
                accepted_backends = []
                for backend in backends:
                    try:
                        if filters(backend) is True:
                            accepted_backends.append(backend)
                    except Exception:  # pylint: disable=broad-except
                        pass
                backends = accepted_backends
            else:
                raise QISKitError('backend filters must be either dict or callable.')

        return backends

    def grouped_backend_names(self):
        """
        Aggregate group names from all providers.

        Returns:
            dict[str: list[str]]: aggregated group dictionary

        Raises:
            ValueError: if a backend is mapped to multiple groups
        """
        groups = {}
        for provider in self.providers:
            groups.update(provider.grouped_backend_names())
        for pair in combinations(groups.values(), r=2):
            if not set.isdisjoint(set(pair[0]), set(pair[1])):
                raise ValueError('duplicate backend group definition')

        return groups

    def deprecated_backend_names(self):
        """
        Aggregate deprecated names from all providers.

        Returns:
            dict[str: str]: aggregated dictionary of deprecated names
        """
        deprecates = {}
        for provider in self.providers:
            deprecates.update(provider.deprecated_backend_names())

        return deprecates

    def aliased_backend_names(self):
        """
        Aggregate aliase names from all providers.

        Returns:
            dict[str: str]: aggregated alias dictionary
        """
        aliases = {}
        for provider in self.providers:
            aliases.update(provider.aliased_backend_names())

        return aliases

    def add_provider(self, provider):
        """
        Add a new provider to the list of known providers.

        Args:
            provider (BaseProvider): Provider instance.

        Returns:
            BaseProvider: the provider instance.

        Raises:
            QISKitError: if trying to add a provider identical to one already registered
        """
        # Check for backend name clashes, emitting a warning.
        current_backends = {str(backend) for backend in self.available_backends()}
        added_backends = {str(backend) for backend in provider.available_backends()}
        common_backends = added_backends.intersection(current_backends)

        # checks for equality of provider instances, based on the __eq__ method
        if provider not in self.providers:
            self.providers.append(provider)
        else:
            raise QISKitError("The same provider has already been registered!")

        if common_backends:
            logger.warning(
                'The backend names "%s" of this provider are already in use. '
                'Refer to documentation for `available_backends()` and `unregister()`.',
                list(common_backends))

        return provider

    def remove_provider(self, provider):
        """
        Remove a provider from the list of known providers.

        Args:
            provider (BaseProvider): provider to be removed.

        Raises:
            QISKitError: if the provider is not registered.
        """
        if isinstance(provider, LocalProvider):
            raise QISKitError("Cannot unregister 'local' provider.")
        try:
            self.providers.remove(provider)
        except ValueError:
            raise QISKitError("'%s' provider is not registered.")

    def resolve_backend_name(self, name):
        """Resolve backend name from a possible short group name, a deprecated name,
        or an alias.

        A group will be resolved in order of member priorities, depending on availability.

        Args:
            name (str): name of backend to resolve

        Returns:
            str: name of resolved backend, which is available from one of the providers

        Raises:
            LookupError: if name cannot be resolved through
            regular available names, nor groups, nor deprecated, nor alias names
        """
        resolved_name = ""
        available = [b.name for b in self.available_backends(filters=None)]
        grouped = self.grouped_backend_names()
        deprecated = self.deprecated_backend_names()
        aliased = self.aliased_backend_names()

        if name in available:
            resolved_name = name
        elif name in grouped:
            available_members = [b for b in grouped[name] if b in available]
            if available_members:
                resolved_name = available_members[0]
        elif name in deprecated:
            resolved_name = deprecated[name]
            logger.warning('WARNING: %s is deprecated. Use %s.', name, resolved_name)
        elif name in aliased:
            resolved_name = aliased[name]

        if resolved_name not in available:
            raise LookupError('backend "{}" not found.'.format(name))

        return resolved_name



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Node for an OPENQASM real number.
"""
from sympy import latex, pi
from sympy.printing.ccode import ccode
from ._node import Node


class Real(Node):
    """Node for an OPENQASM real number.

    This node has no children. The data is in the value field.
    """

    def __init__(self, id):
        """Create the real node."""
        # pylint: disable=redefined-builtin
        Node.__init__(self, "real", None, None)
        self.value = id

    def to_string(self, indent):
        """Print with indent."""
        ind = indent * ' '
        print(ind, 'real', self.value)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        if self.value == pi:
            return "pi"

        return ccode(self.value, precision=prec)

    def latex(self, prec=15, nested_scope=None):
        """Return the corresponding math mode latex string."""
        # pylint: disable=unused-argument
        return latex(self.value)

    def sym(self, nested_scope=None):
        """Return the correspond symbolic number."""
        # pylint: disable=unused-argument
        return self.value

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        # pylint: disable=unused-argument
        return float(self.value.evalf())



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""This module implements the abstract base class for backend jobs

When creating a new backend module it is also necessary to implement this
job interface.
"""

from abc import ABC, abstractmethod


class BaseJob(ABC):
    """Class to handle asynchronous jobs"""

    @abstractmethod
    def __init__(self):
        """Initializes the asynchronous job"""
        pass

    @abstractmethod
    def result(self):
        """Return backend result"""
        pass

    @abstractmethod
    def cancel(self):
        """Attempt to cancel job."""
        pass

    # Property attributes
    #####################
    @property
    @abstractmethod
    def status(self):
        """Get backend status dictionary"""
        pass

    @property
    @abstractmethod
    def running(self):
        """True if job is currently running."""
        pass

    @property
    @abstractmethod
    def done(self):
        """True if call was successfully finished."""
        pass

    @property
    @abstractmethod
    def cancelled(self):
        """True if call was successfully cancelled"""
        pass



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""C++ qasm_simulator gates."""
from .load import load
from .noise import noise
from .save import save
from .snapshot import snapshot
from .wait import wait



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import sys
import os
import platform
from distutils.command.build import build
from multiprocessing import cpu_count
from subprocess import call

from setuptools import setup, find_packages
from setuptools.dist import Distribution


requirements = [
    "IBMQuantumExperience>=1.9.6",
    "matplotlib>=2.1",
    "networkx>=2.0",
    "numpy>=1.13",
    "ply>=3.10",
    "scipy>=0.19",
    "sympy>=1.0",
    "pillow>=4.2.1"
]


# C++ components compilation
class QasmSimulatorCppBuild(build):
    def run(self):
        super().run()
        # Store the current working directory, as invoking cmake involves
        # an out of source build and might interfere with the rest of the steps.
        current_directory = os.getcwd()

        try:
            supported_platforms = ['Linux', 'Darwin', 'Windows']
            current_platform = platform.system()
            if current_platform not in supported_platforms:
                # TODO: stdout is silenced by pip if the full setup.py invocation is
                # successful, unless using '-v' - hence the warnings are not printed.
                print('WARNING: QISKit cpp simulator is meant to be built with these '
                      'platforms: {}. We will support other platforms soon!'
                      .format(supported_platforms))
                return

            cmd_cmake = ['cmake', '-vvv']
            if 'USER_LIB_PATH' in os.environ:
                cmd_cmake.append('-DUSER_LIB_PATH={}'.format(os.environ['USER_LIB_PATH']))
            if current_platform == 'Windows':
                # We only support MinGW so far
                cmd_cmake.append("-GMinGW Makefiles")
            cmd_cmake.append('..')

            cmd_make = ['make', 'pypi_package_copy_qasm_simulator_cpp']

            try:
                cmd_make.append('-j%d' % cpu_count())
            except NotImplementedError:
                print('WARNING: Unable to determine number of CPUs. Using single threaded make.')

            def compile_simulator():
                self.mkpath('out')
                os.chdir('out')
                call(cmd_cmake)
                call(cmd_make)

            self.execute(compile_simulator, [], 'Compiling C++ QASM Simulator')
        except Exception as e:
            print(str(e))
            print("WARNING: Seems like the cpp simulator can't be built, Qiskit will "
                  "install anyway, but won't have this simulator support.")

        # Restore working directory.
        os.chdir(current_directory)


# This is for creating wheel specific platforms
class BinaryDistribution(Distribution):
    def has_ext_modules(self):
        return True


setup(
    name="qiskit",
    version="0.5.7",
    description="Software for developing quantum computing programs",
    long_description="""QISKit is a software development kit for writing
        quantum computing experiments, programs, and applications. Works with
        Python 3.5 and 3.6""",
    url="https://github.com/QISKit/qiskit-terra",
    author="QISKit Development Team",
    author_email="qiskit@us.ibm.com",
    license="Apache 2.0",
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
    ],
    keywords="qiskit sdk quantum",
    packages=find_packages(exclude=['test*']),
    install_requires=requirements,
    include_package_data=True,
    python_requires=">=3.5",
    cmdclass={
        'build': QasmSimulatorCppBuild,
    },
    distclass=BinaryDistribution
)



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Node for an OPENQASM reset statement.
"""
from ._node import Node


class Reset(Node):
    """Node for an OPENQASM reset statement.

    children[0] is a primary node (id or indexedid)
    """

    def __init__(self, children):
        """Create the reset node."""
        Node.__init__(self, 'reset', children, None)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        return "reset " + self.children[0].qasm(prec) + ";"



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""
S=diag(1,i) Clifford phase gate or its inverse.
"""
from qiskit import CompositeGate
from qiskit import InstructionSet
from qiskit import QuantumCircuit
from qiskit import QuantumRegister
from qiskit.qasm import pi
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class SGate(CompositeGate):
    """S=diag(1,i) Clifford phase gate or its inverse."""

    def __init__(self, qubit, circ=None):
        """Create new S gate."""
        super().__init__("s", [], [qubit], circ)
        self.u1(pi / 2, qubit)

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.s(self.arg[0]))

    def qasm(self):
        """Return OPENQASM string."""
        qubit = self.data[0].arg[0]
        phi = self.data[0].param[0]
        if phi > 0:
            return self.data[0]._qasmif("s %s[%d];" % (qubit[0].name, qubit[1]))

        return self.data[0]._qasmif("sdg %s[%d];" % (qubit[0].name, qubit[1]))


def s(self, q):
    """Apply S to q."""
    if isinstance(q, QuantumRegister):
        instructions = InstructionSet()
        for j in range(q.size):
            instructions.add(self.s((q, j)))
        return instructions

    self._check_qubit(q)
    return self._attach(SGate(q, self))


def sdg(self, q):
    """Apply Sdg to q."""
    return self.s(q).inverse()


QuantumCircuit.s = s
QuantumCircuit.sdg = sdg
CompositeGate.s = s
CompositeGate.sdg = sdg



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Tools for compiling a batch of quantum circuits."""
import logging
import copy
import uuid

import numpy as np
import scipy.sparse as sp
import scipy.sparse.csgraph as cs

from qiskit.transpiler._transpilererror import TranspilerError
from qiskit._qiskiterror import QISKitError
from qiskit._quantumcircuit import QuantumCircuit
from qiskit.dagcircuit import DAGCircuit
from qiskit.unroll import DagUnroller, DAGBackend, JsonBackend
from qiskit.mapper import (Coupling, optimize_1q_gates, coupling_list2dict, swap_mapper,
                           cx_cancellation, direction_mapper)
from qiskit._gate import Gate

logger = logging.getLogger(__name__)


# pylint: disable=redefined-builtin
def compile(circuits, backend,
            config=None, basis_gates=None, coupling_map=None, initial_layout=None,
            shots=1024, max_credits=10, seed=None, qobj_id=None, hpc=None,
            pass_manager=None):
    """Compile a list of circuits into a qobj.

    Args:
        circuits (QuantumCircuit or list[QuantumCircuit]): circuits to compile
        backend (BaseBackend): a backend to compile for
        config (dict): dictionary of parameters (e.g. noise) used by runner
        basis_gates (str): comma-separated basis gate set to compile to
        coupling_map (list): coupling map (perhaps custom) to target in mapping
        initial_layout (list): initial layout of qubits in mapping
        shots (int): number of repetitions of each circuit, for sampling
        max_credits (int): maximum credits to use
        seed (int): random seed for simulators
        qobj_id (int): identifier for the generated qobj
        hpc (dict): HPC simulator parameters
        pass_manager (PassManager): a pass_manager for the transpiler stage

    Returns:
        obj: the qobj to be run on the backends

    Raises:
        TranspilerError: in case of bad compile options, e.g. the hpc options.
    """
    if isinstance(circuits, QuantumCircuit):
        circuits = [circuits]

    backend_conf = backend.configuration
    backend_name = backend_conf['name']
    # FIXME (new_backend_names): the API uses the old backend names, but in
    # qiskit we are using the new ones. This makes the qobj names valid (old).
    if getattr(backend, '_name_for_api', None):
        backend_name = backend._name_for_api

    qobj = {}

    # step 1: populate the qobj-level `id`
    qobj_id = qobj_id or str(uuid.uuid4())
    qobj['id'] = qobj_id

    # step 2: populate the qobj-level `config`
    qobj['config'] = {'max_credits': max_credits,
                      'shots': shots,
                      'backend_name': backend_name}

    if hpc is not None and \
            not all(key in hpc for key in ('multi_shot_optimization', 'omp_num_threads')):
        raise TranspilerError('Unknown HPC parameter format!')

    # step 3: populate the `circuits` in qobj, after compiling each circuit
    qobj['circuits'] = []
    if not basis_gates:
        basis_gates = backend_conf['basis_gates']
    if not coupling_map:
        coupling_map = backend_conf['coupling_map']

    for circuit in circuits:
        job = {}

        # step 1: populate the circuit-level `name`
        job["name"] = circuit.name

        # step 2: populate the circuit-level `config`
        if config is None:
            config = {}
        job["config"] = copy.deepcopy(config)
        # TODO: A better solution is to have options to enable/disable optimizations
        num_qubits = sum((len(qreg) for qreg in circuit.get_qregs().values()))
        if num_qubits == 1 or coupling_map == "all-to-all":
            coupling_map = None
        job["config"]["coupling_map"] = coupling_map
        job["config"]["basis_gates"] = basis_gates
        job["config"]["seed"] = seed

        # step 3: populate the circuit `instructions` after compilation
        # step 3a: circuit -> dag
        dag_circuit = DAGCircuit.fromQuantumCircuit(circuit)

        # TODO: move this inside the mapper pass
        # pick a good initial layout if coupling_map is not already satisfied
        # otherwise keep it as q[i]->q[i]
        if (initial_layout is None and
                not backend_conf['simulator'] and
                not _matches_coupling_map(circuit.data, coupling_map)):
            initial_layout = _pick_best_layout(backend, num_qubits, circuit.get_qregs())

        # step 3b: transpile (dag -> dag)
        dag_circuit, final_layout = transpile(
            dag_circuit,
            basis_gates=basis_gates,
            coupling_map=coupling_map,
            initial_layout=initial_layout,
            get_layout=True,
            seed=seed,
            pass_manager=pass_manager)

        # step 3c: dag -> json
        # TODO: populate the Qobj object when Qobj class exists
        # the compiled circuit to be run saved as a dag
        # we assume that transpile() has already expanded gates
        # to the target basis, so we just need to generate json
        list_layout = [[k, v] for k, v in final_layout.items()] if final_layout else None
        job["config"]["layout"] = list_layout
        json_circuit = DagUnroller(dag_circuit, JsonBackend(dag_circuit.basis)).execute()
        job["compiled_circuit"] = json_circuit

        # set eval_symbols=True to evaluate each symbolic expression
        # TODO after transition to qobj, we can drop this
        job["compiled_circuit_qasm"] = dag_circuit.qasm(qeflag=True,
                                                        eval_symbols=True)

        # add job to the qobj
        qobj["circuits"].append(job)

    return qobj


# pylint: disable=redefined-builtin
def transpile(dag_circuit, basis_gates='u1,u2,u3,cx,id', coupling_map=None,
              initial_layout=None, get_layout=False,
              format='dag', seed=None, pass_manager=None):
    """Transform a dag circuit into another dag circuit (transpile), through
    consecutive passes on the dag.

    Args:
        dag_circuit (DAGCircuit): dag circuit to transform via transpilation
        basis_gates (str): a comma seperated string for the target basis gates
        coupling_map (list): A graph of coupling::

            [
             [control0(int), target0(int)],
             [control1(int), target1(int)],
            ]

            eg. [[0, 2], [1, 2], [1, 3], [3, 4]}

        initial_layout (dict): A mapping of qubit to qubit::

                              {
                                ("q", start(int)): ("q", final(int)),
                                ...
                              }
                              eg.
                              {
                                ("q", 0): ("q", 0),
                                ("q", 1): ("q", 1),
                                ("q", 2): ("q", 2),
                                ("q", 3): ("q", 3)
                              }
        get_layout (bool): flag for returning the layout
        format (str): The target format of the compilation:
            {'dag', 'json', 'qasm'}
        seed (int): random seed for simulators
        pass_manager (PassManager): pass manager instance for the tranpilation process
            If None, a default set of passes are run.
            Otherwise, the passes defined in it will run.
            If contains no passes in it, no dag transformations occur.

    Returns:
        object: If get_layout == False, the compiled circuit in the specified
            format. If get_layout == True, a tuple is returned, with the
            second element being the layout.

    Raises:
        TranspilerError: if the format is not valid.
    """
    final_layout = None

    if pass_manager:
        # run the passes specified by the pass manager
        for pass_ in pass_manager.passes():
            pass_.run(dag_circuit)
    else:
        # default set of passes
        # TODO: move each step here to a pass, and use a default passmanager below
        basis = basis_gates.split(',') if basis_gates else []
        dag_unroller = DagUnroller(dag_circuit, DAGBackend(basis))
        dag_circuit = dag_unroller.expand_gates()
        # if a coupling map is given compile to the map
        if coupling_map:
            logger.info("pre-mapping properties: %s",
                        dag_circuit.property_summary())
            # Insert swap gates
            coupling = Coupling(coupling_list2dict(coupling_map))
            logger.info("initial layout: %s", initial_layout)
            dag_circuit, final_layout = swap_mapper(
                dag_circuit, coupling, initial_layout, trials=20, seed=seed)
            logger.info("final layout: %s", final_layout)
            # Expand swaps
            dag_unroller = DagUnroller(dag_circuit, DAGBackend(basis))
            dag_circuit = dag_unroller.expand_gates()
            # Change cx directions
            dag_circuit = direction_mapper(dag_circuit, coupling)
            # Simplify cx gates
            cx_cancellation(dag_circuit)
            # Simplify single qubit gates
            dag_circuit = optimize_1q_gates(dag_circuit)
            logger.info("post-mapping properties: %s",
                        dag_circuit.property_summary())

    # choose output format
    # TODO: do we need all of these formats, or just the dag?
    if format == 'dag':
        compiled_circuit = dag_circuit
    elif format == 'json':
        # FIXME: JsonBackend is wrongly taking an ordered dict as basis, not list
        dag_unroller = DagUnroller(dag_circuit, JsonBackend(dag_circuit.basis))
        compiled_circuit = dag_unroller.execute()
    elif format == 'qasm':
        compiled_circuit = dag_circuit.qasm()
    else:
        raise TranspilerError('unrecognized circuit format')

    if get_layout:
        return compiled_circuit, final_layout
    return compiled_circuit


def _best_subset(backend, n_qubits):
    """Computes the qubit mapping with the best
    connectivity.

    Parameters:
        backend (Qiskit.BaseBackend): A QISKit backend instance.
        n_qubits (int): Number of subset qubits to consider.

    Returns:
        ndarray: Array of qubits to use for best
                connectivity mapping.

    Raises:
        QISKitError: Wrong number of qubits given.
    """
    if n_qubits == 1:
        return np.array([0])
    elif n_qubits <= 0:
        raise QISKitError('Number of qubits <= 0.')

    device_qubits = backend.configuration['n_qubits']
    if n_qubits > device_qubits:
        raise QISKitError('Number of qubits greater than device.')

    cmap = np.asarray(backend.configuration['coupling_map'])
    data = np.ones_like(cmap[:, 0])
    sp_cmap = sp.coo_matrix((data, (cmap[:, 0], cmap[:, 1])),
                            shape=(device_qubits, device_qubits)).tocsr()
    best = 0
    best_map = None
    # do bfs with each node as starting point
    for k in range(sp_cmap.shape[0]):
        bfs = cs.breadth_first_order(sp_cmap, i_start=k, directed=False,
                                     return_predecessors=False)

        connection_count = 0
        for i in range(n_qubits):
            node_idx = bfs[i]
            for j in range(sp_cmap.indptr[node_idx],
                           sp_cmap.indptr[node_idx + 1]):
                node = sp_cmap.indices[j]
                for counter in range(n_qubits):
                    if node == bfs[counter]:
                        connection_count += 1
                        break

        if connection_count > best:
            best = connection_count
            best_map = bfs[0:n_qubits]
    return best_map


def _matches_coupling_map(instructions, coupling_map):
    """Iterate over circuit instructions to check if all multi-qubit couplings
    match the qubit coupling graph in the backend.

    Parameters:
            instructions (list): List of circuit instructions.
            coupling_map (list): Backend coupling map, represented as an adjacency list.

    Returns:
            True: If all instructions readily fit the backend coupling graph.

            False: If there's at least one instruction that uses multiple qubits
                   which does not match the backend couplings.
    """
    for instruction in instructions:
        if isinstance(instruction, Gate) and instruction.is_multi_qubit():
            if instruction.get_qubit_coupling() not in coupling_map:
                return False
    return True


def _pick_best_layout(backend, num_qubits, qregs):
    """ Pick a convenient layout depending on the best matching qubit connectivity

    Parameters:
        backend (BaseBackend) : The backend with the coupling_map for searching
        num_qubits (int): Number of qubits
        qregs (list): The list of quantum registers

    Returns:
        initial_layout: A special ordered layout

    """
    best_sub = _best_subset(backend, num_qubits)
    layout = {}
    map_iter = 0
    for key, value in qregs.items():
        for i in range(value.size):
            layout[(key, i)] = ('q', best_sub[map_iter])
            map_iter += 1
    return layout



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""This module defines an enumerated type for the state of backend jobs"""

import enum


class JobStatus(enum.Enum):
    """Class for job status enumerated type."""
    INITIALIZING = 'job is being initialized'
    QUEUED = 'job is queued'
    RUNNING = 'job is actively running'
    CANCELLED = 'job has been cancelled'
    DONE = 'job has successfully run'
    ERROR = 'job incurred error'



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Quantum circuit object.
"""
import itertools
from collections import OrderedDict
from ._qiskiterror import QISKitError
from ._register import Register
from ._quantumregister import QuantumRegister
from ._classicalregister import ClassicalRegister
from ._measure import Measure
from ._reset import Reset
from ._instructionset import InstructionSet


class QuantumCircuit(object):
    """Quantum circuit."""
    instances = 0
    prefix = 'circuit'

    # Class variable OPENQASM header
    header = "OPENQASM 2.0;"

    # Class variable with gate definitions
    # This is a dict whose values are dicts with the
    # following keys:
    #   "print" = True or False
    #   "opaque" = True or False
    #   "n_args" = number of real parameters
    #   "n_bits" = number of qubits
    #   "args"   = list of parameter names
    #   "bits"   = list of qubit names
    #   "body"   = GateBody AST node
    definitions = OrderedDict()

    def __init__(self, *regs, name=None):
        """Create a new circuit.

        Args:
            *regs (Registers): registers to include in the circuit.
            name (str or None): the name of the quantum circuit. If
                None, an automatically generated identifier will be
                assigned.

        Raises:
            QISKitError: if the circuit name, if given, is not valid.
        """
        self._increment_instances()
        if name is None:
            name = self.cls_prefix() + str(self.cls_instances())

        if not isinstance(name, str):
            raise QISKitError("The circuit name should be a string "
                              "(or None for autogenerate a name).")

        self.name = name
        # Data contains a list of instructions in the order they were applied.
        self.data = []
        # This is a map of registers bound to this circuit, by name.
        self.regs = OrderedDict()
        self.add(*regs)

    @classmethod
    def _increment_instances(cls):
        cls.instances += 1

    @classmethod
    def cls_instances(cls):
        """Return the current number of instances of this class,
        useful for auto naming."""
        return cls.instances

    @classmethod
    def cls_prefix(cls):
        """Return the prefix to use for auto naming."""
        return cls.prefix

    def has_register(self, register):
        """
        Test if this circuit has the register r.

        Return True or False.
        """
        if register.name in self.regs:
            registers = self.regs[register.name]
            if registers.size == register.size:
                if ((isinstance(register, QuantumRegister) and
                     isinstance(registers, QuantumRegister)) or
                        (isinstance(register, ClassicalRegister) and
                         isinstance(registers, ClassicalRegister))):
                    return True
        return False

    def get_qregs(self):
        """Get the qregs from the registers."""
        qregs = OrderedDict()
        for name, register in self.regs.items():
            if isinstance(register, QuantumRegister):
                qregs[name] = register
        return qregs

    def get_cregs(self):
        """Get the cregs from the registers."""
        cregs = OrderedDict()
        for name, register in self.regs.items():
            if isinstance(register, ClassicalRegister):
                cregs[name] = register
        return cregs

    def combine(self, rhs):
        """
        Append rhs to self if self contains compatible registers.

        Two circuits are compatible if they contain the same registers
        or if they contain different registers with unique names. The
        returned circuit will contain all unique registers between both
        circuits.

        Return self + rhs as a new object.
        """
        combined_registers = []
        # Check registers in LHS are compatible with RHS
        for name, register in self.regs.items():
            if name in rhs.regs and register != rhs.regs[name]:
                raise QISKitError("circuits are not compatible")
            else:
                combined_registers.append(register)
        # Add registers in RHS not in LHS
        complement_registers = set(rhs.regs) - set(self.regs)
        for name in complement_registers:
            combined_registers.append(rhs.regs[name])
        # Make new circuit with combined registers
        circuit = QuantumCircuit(*combined_registers)
        for gate in itertools.chain(self.data, rhs.data):
            gate.reapply(circuit)
        return circuit

    def extend(self, rhs):
        """
        Append rhs to self if self if it contains compatible registers.

        Two circuits are compatible if they contain the same registers
        or if they contain different registers with unique names. The
        returned circuit will contain all unique registers between both
        circuits.

        Modify and return self.
        """
        # Check compatibility and add new registers
        for name, register in rhs.regs.items():
            if name not in self.regs:
                self.add(register)
            elif name in self.regs and register != self.regs[name]:
                raise QISKitError("circuits are not compatible")

        # Add new gates
        for gate in rhs.data:
            gate.reapply(self)
        return self

    def __add__(self, rhs):
        """Overload + to implement self.concatenate."""
        return self.combine(rhs)

    def __iadd__(self, rhs):
        """Overload += to implement self.extend."""
        return self.extend(rhs)

    def __len__(self):
        """Return number of operations in circuit."""
        return len(self.data)

    def __getitem__(self, item):
        """Return indexed operation."""
        return self.data[item]

    def _attach(self, gate):
        """Attach a gate."""
        self.data.append(gate)
        return gate

    def add(self, *regs):
        """Add registers."""
        for register in regs:
            if not isinstance(register, Register):
                raise QISKitError("expected a register")
            if register.name not in self.regs:
                self.regs[register.name] = register
            else:
                raise QISKitError("register name \"%s\" already exists"
                                  % register.name)

    def _check_qreg(self, register):
        """Raise exception if r is not in this circuit or not qreg."""
        if not isinstance(register, QuantumRegister):
            raise QISKitError("expected quantum register")
        if not self.has_register(register):
            raise QISKitError(
                "register '%s' not in this circuit" %
                register.name)

    def _check_qubit(self, qubit):
        """Raise exception if qubit is not in this circuit or bad format."""
        if not isinstance(qubit, tuple):
            raise QISKitError("%s is not a tuple."
                              "A qubit should be formated as a tuple." % str(qubit))
        if not len(qubit) == 2:
            raise QISKitError("%s is not a tuple with two elements, but %i instead" % len(qubit))
        if not isinstance(qubit[1], int):
            raise QISKitError("The second element of a tuple defining a qubit should be an int:"
                              "%s was found instead" % type(qubit[1]).__name__)
        self._check_qreg(qubit[0])
        qubit[0].check_range(qubit[1])

    def _check_creg(self, register):
        """Raise exception if r is not in this circuit or not creg."""
        if not isinstance(register, ClassicalRegister):
            raise QISKitError("expected classical register")
        if not self.has_register(register):
            raise QISKitError(
                "register '%s' not in this circuit" %
                register.name)

    def _check_dups(self, qubits):
        """Raise exception if list of qubits contains duplicates."""
        squbits = set(qubits)
        if len(squbits) != len(qubits):
            raise QISKitError("duplicate qubit arguments")

    def _gate_string(self, name):
        """Return a QASM string for the named gate."""
        out = ""
        if self.definitions[name]["opaque"]:
            out = "opaque " + name
        else:
            out = "gate " + name
        if self.definitions[name]["n_args"] > 0:
            out += "(" + ",".join(self.definitions[name]["args"]) + ")"
        out += " " + ",".join(self.definitions[name]["bits"])
        if self.definitions[name]["opaque"]:
            out += ";"
        else:
            out += "\n{\n" + self.definitions[name]["body"].qasm() + "}\n"
        return out

    def qasm(self):
        """Return OPENQASM string."""
        string_temp = self.header + "\n"
        for gate_name in self.definitions:
            if self.definitions[gate_name]["print"]:
                string_temp += self._gate_string(gate_name)
        for register in self.regs.values():
            string_temp += register.qasm() + "\n"
        for instruction in self.data:
            string_temp += instruction.qasm() + "\n"
        return string_temp

    def measure(self, qubit, cbit):
        """Measure quantum bit into classical bit (tuples).

        Returns:
            Gate: the attached measure gate.

        Raises:
            QISKitError: if qubit is not in this circuit or bad format;
                if cbit is not in this circuit or not creg.
        """
        if isinstance(qubit, QuantumRegister) and \
           isinstance(cbit, ClassicalRegister) and len(qubit) == len(cbit):
            instructions = InstructionSet()
            for i in range(qubit.size):
                instructions.add(self.measure((qubit, i), (cbit, i)))
            return instructions

        self._check_qubit(qubit)
        self._check_creg(cbit[0])
        cbit[0].check_range(cbit[1])
        return self._attach(Measure(qubit, cbit, self))

    def reset(self, quantum_register):
        """Reset q."""
        if isinstance(quantum_register, QuantumRegister):
            instructions = InstructionSet()
            for sizes in range(quantum_register.size):
                instructions.add(self.reset((quantum_register, sizes)))
            return instructions
        self._check_qubit(quantum_register)
        return self._attach(Reset(quantum_register, self))



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=wrong-import-order
# pylint: disable=redefined-builtin

"""Main QISKit public functionality."""

# First, check for required Python and API version
from . import _util

from ._qiskiterror import QISKitError
from ._classicalregister import ClassicalRegister
from ._quantumregister import QuantumRegister
from ._quantumcircuit import QuantumCircuit
from ._gate import Gate
from ._compositegate import CompositeGate
from ._instruction import Instruction
from ._instructionset import InstructionSet
from ._reset import Reset
from ._measure import Measure

# The qiskit.extensions.x imports needs to be placed here due to the
# mechanism for adding gates dynamically.
import qiskit.extensions.standard
import qiskit.extensions.quantum_initializer

from ._quantumprogram import QuantumProgram
from ._result import Result

from .wrapper._wrapper import (
    available_backends, local_backends, remote_backends,
    get_backend, compile, execute, register, unregister,
    registered_providers, load_qasm_string, load_qasm_file, least_busy)

# Import the wrapper, to make it available when doing "import qiskit".
from . import wrapper

import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(ROOT_DIR, "VERSION.txt"), "r") as version_file:
    __version__ = version_file.read().strip()



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Node for an OPENQASM custom gate body.
"""
from ._node import Node


class GateBody(Node):
    """Node for an OPENQASM custom gate body.

    children is a list of gate operation nodes.
    These are one of barrier, custom_unitary, U, or CX.
    """

    def __init__(self, children):
        """Create the gatebody node."""
        Node.__init__(self, 'gate_body', children, None)

    def qasm(self, prec=15):
        """Return the corresponding OPENQASM string."""
        string = ""
        for children in self.children:
            string += "  " + children.qasm(prec) + "\n"
        return string

    def calls(self):
        """Return a list of custom gate names in this gate body."""
        lst = []
        for children in self.children:
            if children.type == "custom_unitary":
                lst.append(children.name)
        return lst



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Utilities regarding the creation of QuantumCircuits from a variety of different
file formats.
"""

import os
from qiskit import QISKitError
from qiskit.qasm import Qasm
from qiskit.unroll import Unroller, CircuitBackend


def circuit_from_qasm_string(qasm_string, name=None,
                             basis_gates="id,u0,u1,u2,u3,x,y,z,h,s,sdg,t,tdg,"
                             "rx,ry,rz,cx,cy,cz,ch,crz,cu1,cu3,swap,ccx,"
                             "cswap"):

    """Construct a quantum circuit from a qasm representation (string).

    Args:
        qasm_string (str): a string of qasm, or a filename containing qasm.
        basis_gates (str): basis gates for the quantum circuit.
        name (str or None): the name of the quantum circuit after loading
            qasm text into it. If no name given, assign automatically.
    Returns:
        QuantumCircuit: circuit constructed from qasm.
    Raises:
        QISKitError: if the string is not valid QASM
    """

    node_circuit = Qasm(data=qasm_string).parse()
    unrolled_circuit = Unroller(
        node_circuit, CircuitBackend(basis_gates.split(",")))
    circuit_unrolled = unrolled_circuit.execute()
    if name:
        circuit_unrolled.name = name
    return circuit_unrolled


def circuit_from_qasm_file(qasm_file, name=None,
                           basis_gates="id,u0,u1,u2,u3,x,y,z,h,s,sdg,t,tdg,rx,"
                           "ry,rz,cx,cy,cz,ch,crz,cu1,cu3,swap,ccx,"
                           "cswap"):

    """Construct a quantum circuit from a qasm representation (file).

    Args:
        qasm_file (str): a string for the filename including its location.
        name (str or None): the name of the quantum circuit after
            loading qasm text into it. If no name is give the name is of
            the text file.
        basis_gates (str): basis gates for the quantum circuit.
    Returns:
        QuantumCircuit: circuit constructed from qasm.
    Raises:
        QISKitError: if the file cannot be read.
    """
    if not os.path.exists(qasm_file):
        raise QISKitError('qasm file "{0}" not found'.format(qasm_file))
    if not name:
        name = os.path.splitext(os.path.basename(qasm_file))[0]

    with open(qasm_file) as file:
        qasm_data = file.read()

    return circuit_from_qasm_string(
        qasm_data, name=name, basis_gates=basis_gates)



# -*- coding: utf-8 -*-

# Copyright 2018, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Pass for peep-hole cancellation of consecutive CX gates.
"""
from qiskit.transpiler._basepass import BasePass


class CXCancellation(BasePass):
    """Cancel back-to-back 'cx' gates in dag."""

    def __init__(self):
        pass

    def run(self, dag):
        """
        run one pass of cx cancellation on the circuit

        Args:
            dag (DAGCircuit): the directed acyclic graph to run on
        """
        cx_runs = dag.collect_runs(["cx"])
        for cx_run in cx_runs:
            # Partition the cx_run into chunks with equal gate arguments
            partition = []
            chunk = []
            for i in range(len(cx_run) - 1):
                chunk.append(cx_run[i])
                qargs0 = dag.multi_graph.node[cx_run[i]]["qargs"]
                qargs1 = dag.multi_graph.node[cx_run[i + 1]]["qargs"]
                if qargs0 != qargs1:
                    partition.append(chunk)
                    chunk = []
            chunk.append(cx_run[-1])
            partition.append(chunk)
            # Simplify each chunk in the partition
            for chunk in partition:
                if len(chunk) % 2 == 0:
                    for n in chunk:
                        dag._remove_op_node(n)
                else:
                    for n in chunk[1:]:
                        dag._remove_op_node(n)



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Standard gates."""
from .barrier import barrier
from .ccx import ccx
from .cswap import cswap
from .cx import cx
from .cxbase import cx_base
from .cy import cy
from .cz import cz
from .swap import swap
from .h import h
from .iden import iden
from .s import s
from .t import t
from .u0 import u0
from .u1 import u1
from .u2 import u2
from .u3 import u3
from .ubase import u_base
from .x import x
from .y import y
from .z import z
from .rx import rx
from .ry import ry
from .rz import rz
from .cu1 import cu1
from .ch import ch
from .crz import crz
from .cu3 import cu3
from .rzz import rzz



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Simulator command to load a saved quantum state.
"""
from qiskit import Instruction
from qiskit import QuantumCircuit
from qiskit import CompositeGate
from qiskit import QuantumRegister
from qiskit.extensions._extensionerror import ExtensionError
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class Load(Instruction):
    """Simulator load instruction."""

    def __init__(self, slot, qubits, circ):
        """Create new load instruction."""
        super().__init__("load", [slot], list(qubits), circ)

    def inverse(self):
        """Special case. Return self."""
        return self

    def qasm(self):
        """Return OPENQASM string."""
        string = "load(%d) " % self.param[0]
        for j in range(len(self.arg)):
            if len(self.arg[j]) == 1:
                string += "%s" % self.arg[j].name
            else:
                string += "%s[%d]" % (self.arg[j][0].name, self.arg[j][1])
            if j != len(self.arg) - 1:
                string += ","
        string += ";"
        return string

    def reapply(self, circ):
        """Reapply this instruction to corresponding qubits in circ."""
        self._modifiers(circ.load(self.param[0]))


def load(self, slot):
    """Load the internal simulator representation (statevector, probability,
    density matrix, clifford table)
    Works on all qubits, and prevents reordering (like barrier).

    Args:
        slot (int): a slot to load from

    Returns:
        QuantumCircuit: with attached command

    Raises:
        ExtensionError: malformed command
    """
    tuples = []
    if isinstance(self, QuantumCircuit):
        for register in self.regs.values():
            if isinstance(register, QuantumRegister):
                tuples.append(register)
    if not tuples:
        raise ExtensionError("no qubits for load")
    if slot is None:
        raise ExtensionError("no load slot passed")
    qubits = []
    for tuple_element in tuples:
        if isinstance(tuple_element, QuantumRegister):
            for j in range(tuple_element.size):
                self._check_qubit((tuple_element, j))
                qubits.append((tuple_element, j))
        else:
            self._check_qubit(tuple_element)
            qubits.append(tuple_element)
    self._check_dups(qubits)
    return self._attach(Load(slot, qubits, self))


# Add to QuantumCircuit and CompositeGate classes
QuantumCircuit.load = load
CompositeGate.load = load



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""
SWAP gate.
"""
from qiskit import CompositeGate
from qiskit import Gate
from qiskit import QuantumCircuit
from qiskit._instructionset import InstructionSet
from qiskit._quantumregister import QuantumRegister
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class SwapGate(Gate):
    """SWAP gate."""

    def __init__(self, ctl, tgt, circ=None):
        """Create new SWAP gate."""
        super().__init__("swap", [], [ctl, tgt], circ)

    def qasm(self):
        """Return OPENQASM string."""
        ctl = self.arg[0]
        tgt = self.arg[1]
        return self._qasmif("swap %s[%d],%s[%d];" % (ctl[0].name, ctl[1],
                                                     tgt[0].name, tgt[1]))

    def inverse(self):
        """Invert this gate."""
        return self  # self-inverse

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.swap(self.arg[0], self.arg[1]))


def swap(self, ctl, tgt):
    """Apply SWAP from ctl to tgt."""
    if isinstance(ctl, QuantumRegister) and \
            isinstance(tgt, QuantumRegister) and len(ctl) == len(tgt):
        instructions = InstructionSet()
        for j in range(ctl.size):
            instructions.add(self.swap((ctl, j), (tgt, j)))
        return instructions

    self._check_qubit(ctl)
    self._check_qubit(tgt)
    self._check_dups([ctl, tgt])
    return self._attach(SwapGate(ctl, tgt, self))


QuantumCircuit.swap = swap
CompositeGate.swap = swap



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""
Rotation around the z-axis.
"""
from qiskit import CompositeGate
from qiskit import Gate
from qiskit import InstructionSet
from qiskit import QuantumCircuit
from qiskit import QuantumRegister
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class RZGate(Gate):
    """rotation around the z-axis."""

    def __init__(self, phi, qubit, circ=None):
        """Create new rz single qubit gate."""
        super().__init__("rz", [phi], [qubit], circ)

    def qasm(self):
        """Return OPENQASM string."""
        qubit = self.arg[0]
        phi = self.param[0]
        return self._qasmif("rz(%s) %s[%d];" % (phi, qubit[0].name, qubit[1]))

    def inverse(self):
        """Invert this gate.

        rz(phi)^dagger = rz(-phi)
        """
        self.param[0] = -self.param[0]
        return self

    def reapply(self, circ):
        """Reapply this gate to corresponding qubits in circ."""
        self._modifiers(circ.rz(self.param[0], self.arg[0]))


def rz(self, phi, q):
    """Apply Rz to q."""
    if isinstance(q, QuantumRegister):
        instructions = InstructionSet()
        for j in range(q.size):
            instructions.add(self.rz(phi, (q, j)))
        return instructions

    self._check_qubit(q)
    return self._attach(RZGate(phi, q, self))


QuantumCircuit.rz = rz
CompositeGate.rz = rz



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
DAG Unroller
"""

import networkx as nx

from qiskit.unroll import Unroller
from qiskit.qasm._node import Real, Id, IdList, ExpressionList, Gate, \
                              PrimaryList, Int, IndexedId, Qreg, If, Creg, \
                              Program, CustomUnitary
from ._unrollererror import UnrollerError
from ._dagbackend import DAGBackend


class DagUnroller(object):
    """An Unroller that takes Dag circuits as the input."""
    def __init__(self, dag_circuit, backend=None):
        if dag_circuit is None:
            raise UnrollerError('Invalid dag circuit!!')

        self.dag_circuit = dag_circuit
        self.backend = backend

    def set_backend(self, backend):
        """Set the backend object."""
        self.backend = backend

    def execute(self):
        """Interpret OPENQASM and make appropriate backend calls."""
        if self.backend is not None:
            self._process()
            return self.backend.get_output()
        else:
            raise UnrollerError("backend not attached")

    # TODO This method should merge with .execute(), so the output will depend
    # on the backend associated with this DagUnroller instance
    def expand_gates(self, basis=None):
        """Expand all gate nodes to the given basis.

        If basis is empty, each custom gate node is replaced by its
        implementation over U and CX. If basis contains names, then
        those custom gates are not expanded. For example, if "u3"
        is in basis, then the gate "u3" will not be expanded wherever
        it occurs.

        This member function replicates the behavior of the unroller
        module without using the OpenQASM parser.
        """

        if basis is None:
            basis = self.backend.basis

        if not isinstance(self.backend, DAGBackend):
            raise UnrollerError("expand_gates only accepts a DAGBackend!!")

        # Build the Gate AST nodes for user-defined gates
        gatedefs = []
        for name, gate in self.dag_circuit.gates.items():
            children = [Id(name, 0, "")]
            if gate["n_args"] > 0:
                children.append(ExpressionList(list(
                    map(lambda x: Id(x, 0, ""),
                        gate["args"])
                )))
            children.append(IdList(list(
                map(lambda x: Id(x, 0, ""),
                    gate["bits"])
            )))
            children.append(gate["body"])
            gatedefs.append(Gate(children))
        # Walk through the DAG and examine each node
        builtins = ["U", "CX", "measure", "reset", "barrier"]
        topological_sorted_list = list(nx.topological_sort(self.dag_circuit.multi_graph))
        for node in topological_sorted_list:
            current_node = self.dag_circuit.multi_graph.node[node]
            if current_node["type"] == "op" and \
               current_node["name"] not in builtins + basis and \
               not self.dag_circuit.gates[current_node["name"]]["opaque"]:
                subcircuit, wires = self._build_subcircuit(gatedefs,
                                                           basis,
                                                           current_node["name"],
                                                           current_node["params"],
                                                           current_node["qargs"],
                                                           current_node["condition"])
                self.dag_circuit.substitute_circuit_one(node, subcircuit, wires)
        return self.dag_circuit

    def _build_subcircuit(self, gatedefs, basis, gate_name, gate_params, gate_args,
                          gate_condition):
        """Build DAGCircuit for a given user-defined gate node.

        gatedefs = dictionary of Gate AST nodes for user-defined gates
        gate_name = name of gate to expand to target_basis (nd["name"])
        gate_params = list of gate parameters (nd["params"])
        gate_args = list of gate arguments (nd["qargs"])
        gate_condition = None or tuple (string, int) (nd["condition"])

        Returns (subcircuit, wires) where subcircuit is the DAGCircuit
        corresponding to the user-defined gate node expanded to target_basis
        and wires is the list of input wires to the subcircuit in order
        corresponding to the gate's arguments.
        """

        children = [Id(gate_name, 0, "")]
        if gate_params:
            children.append(
                ExpressionList(list(map(Real, gate_params)))
            )
        new_wires = [("q", j) for j in range(len(gate_args))]
        children.append(
            PrimaryList(
                list(map(lambda x: IndexedId(
                    [Id(x[0], 0, ""), Int(x[1])]
                ), new_wires))
            )
        )
        gate_node = CustomUnitary(children)
        id_int = [Id("q", 0, ""), Int(len(gate_args))]
        # Make a list of register declaration nodes
        reg_nodes = [
            Qreg(
                [
                    IndexedId(id_int)
                ]
            )
        ]
        # Add an If node when there is a condition present
        if gate_condition:
            gate_node = If([
                Id(gate_condition[0], 0, ""),
                Int(gate_condition[1]),
                gate_node
            ])
            new_wires += [(gate_condition[0], j)
                          for j in range(self.dag_circuit.cregs[gate_condition[0]])]
            reg_nodes.append(
                Creg([
                    IndexedId([
                        Id(gate_condition[0], 0, ""),
                        Int(self.dag_circuit.cregs[gate_condition[0]])
                    ])
                ])
            )

        # Build the whole program's AST
        sub_ast = Program(gatedefs + reg_nodes + [gate_node])
        # Interpret the AST to give a new DAGCircuit over backend basis
        sub_circuit = Unroller(sub_ast, DAGBackend(basis)).execute()
        return sub_circuit, new_wires

    def _process(self):
        for name, width in self.dag_circuit.qregs.items():
            self.backend.new_qreg(name, width)
        for name, width in self.dag_circuit.cregs.items():
            self.backend.new_creg(name, width)
        for name, data in self.dag_circuit.gates.items():
            self.backend.define_gate(name, data)
        for n in nx.topological_sort(self.dag_circuit.multi_graph):
            current_node = self.dag_circuit.multi_graph.node[n]
            if current_node["type"] == "op":
                params = map(Real, current_node["params"])
                params = list(params)
                if current_node["condition"] is not None:
                    self.backend.set_condition(current_node["condition"][0],
                                               current_node["condition"][1])
                if not current_node["cargs"]:
                    if current_node["name"] == "U":
                        self.backend.u(params, current_node["qargs"][0])
                    elif current_node["name"] == "CX":
                        self.backend.cx(current_node["qargs"][0], current_node["qargs"][1])
                    elif current_node["name"] == "barrier":
                        self.backend.barrier([current_node["qargs"]])
                    elif current_node["name"] == "reset":
                        self.backend.reset(current_node["qargs"][0])
                    else:
                        self.backend.start_gate(current_node["name"], params,
                                                current_node["qargs"])
                        self.backend.end_gate(current_node["name"], params, current_node["qargs"])
                else:
                    if current_node["name"] == "measure":
                        if len(current_node["cargs"]) != 1 or len(current_node["qargs"]) != 1 \
                           or current_node["params"]:
                            raise UnrollerError("Bad node data!!")

                        self.backend.measure(current_node["qargs"][0], current_node["cargs"][0])
                    else:
                        raise UnrollerError("Bad node data!")

                self.backend.drop_condition()
        return self.backend.get_output()



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# pylint: disable=invalid-name

"""Gates tools."""

from qiskit import InstructionSet
from qiskit import QuantumRegister
from qiskit.extensions.standard import header  # pylint: disable=unused-import


def attach_gate(element, quantum_register, gate, gate_class):
    """Attach a gate."""
    if isinstance(quantum_register, QuantumRegister):
        gs = InstructionSet()
        for _ in range(quantum_register.size):
            gs.add(gate)
        return gs

    element._check_qubit(quantum_register)
    return element._attach(gate_class)



# -*- coding: utf-8 -*-

# Copyright 2017, IBM.
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Simulator command to snapshot internal simulator representation.
"""
from qiskit import Instruction
from qiskit import QuantumCircuit
from qiskit import CompositeGate
from qiskit import QuantumRegister
from qiskit.extensions._extensionerror import ExtensionError
from qiskit.extensions.standard import header  # pylint: disable=unused-import


class Snapshot(Instruction):
    """Simulator snapshot instruction."""

    def __init__(self, slot, qubits, circ):
        """Create new snapshot instruction."""
        super().__init__("snapshot", [slot], list(qubits), circ)

    def inverse(self):
        """Special case. Return self."""
        return self

    def qasm(self):
        """Return OPENQASM string."""
        string = "snapshot(%d) " % self.param[0]
        for j in range(len(self.arg)):
            if len(self.arg[j]) == 1:
                string += "%s" % self.arg[j].name
            else:
                string += "%s[%d]" % (self.arg[j][0].name, self.arg[j][1])
            if j != len(self.arg) - 1:
                string += ","
        string += ";"
        return string

    def reapply(self, circ):
        """Reapply this instruction to corresponding qubits in circ."""
        self._modifiers(circ.snapshot(self.param[0]))


def snapshot(self, slot):
    """Take a snapshot of the internal simulator representation (statevector,
    probability, density matrix, clifford table)
    Works on all qubits, and prevents reordering (like barrier).

    Args:
        slot (int): a snapshot slot to report the result

    Returns:
        QuantumCircuit: with attached command

    Raises:
        ExtensionError: malformed command
    """
    tuples = []
    if isinstance(self, QuantumCircuit):
        for register in self.regs.values():
            if isinstance(register, QuantumRegister):
                tuples.append(register)
    if not tuples:
        raise ExtensionError("no qubits for snapshot")
    if slot is None:
        raise ExtensionError("no snapshot slot passed")
    qubits = []
    for tuple_element in tuples:
        if isinstance(tuple_element, QuantumRegister):
            for j in range(tuple_element.size):
                self._check_qubit((tuple_element, j))
                qubits.append((tuple_element, j))
        else:
            self._check_qubit(tuple_element)
            qubits.append(tuple_element)
    self._check_dups(qubits)
    return self._attach(Snapshot(slot, qubits, self))


# Add to QuantumCircuit and CompositeGate classes
QuantumCircuit.snapshot = snapshot
CompositeGate.snapshot = snapshot
